import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
from contextlib import suppress
from ks_includes.screen_panel import ScreenPanel
from ks_includes.widgets.heatergraph import HeaterGraph
from ks_includes.widgets.keypad import Keypad
from ks_includes.KlippyGtk import find_widget


class Panel(ScreenPanel):
    graph_update = None
    active_heater = None

    def __init__(self, screen, title, **kwargs):
        title = title or _("Temperature")
        super().__init__(screen, title)
        self.left_panel = None
        self.devices = {}
        self.popover = Gtk.Popover(position=Gtk.PositionType.BOTTOM)
        self.popover_buttons = {}
        self.long_press = {}
        self.popover_device = None
        self.h = self.f = 0
        self.tempdeltas = ["1", "5", "10", "25"]
        self.tempdelta = self.tempdeltas[-2]
        self.show_preheat = self._printer.state not in ("printing", "paused")
        self.preheat_options = self._screen._config.get_preheat_options()
        self.grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        self._gtk.reset_temp_color()
        self.extra_selection = None
        self.numpad_visible = False

        if self._screen.vertical_mode:
            self.grid.attach(self.create_left_panel(), 0, 0, 1, 3)
            self.grid.attach(self.create_right_panel(), 0, 3, 1, 2)
        else:
            self.grid.attach(self.create_left_panel(), 0, 0, 1, 1)
            self.grid.attach(self.create_right_panel(), 1, 0, 1, 1)

        self.content.add(self.grid)

    def create_right_panel(self):
        cooldown = self._gtk.Button(
            "cool-down", _("Cooldown"), "color4", self.bts, Gtk.PositionType.LEFT, 1
        )
        adjust = self._gtk.Button(
            "fine-tune", None, "color3", self.bts * 1.4, Gtk.PositionType.LEFT, 1
        )
        cooldown.connect("clicked", self.set_temperature, "cooldown")
        adjust.connect("clicked", self.switch_preheat_adjust)

        right = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        right.attach(cooldown, 0, 0, 2, 1)
        right.attach(adjust, 2, 0, 1, 1)
        if self.show_preheat:
            right.attach(self.preheat(), 0, 1, 3, 2)
        else:
            right.attach(self.delta_adjust(), 0, 1, 3, 2)
        return right

    def switch_preheat_adjust(self, widget):
        self.show_preheat ^= True
        if self._screen.vertical_mode:
            row = self.grid.get_child_at(0, 3)
            self.grid.remove(row)
            self.grid.attach(self.create_right_panel(), 0, 3, 1, 2)
        else:
            self.grid.remove_column(1)
            self.grid.attach(self.create_right_panel(), 1, 0, 1, 1)
        self.grid.show_all()

    def preheat(self):
        self.labels["preheat_grid"] = Gtk.Grid(
            row_homogeneous=True, column_homogeneous=True
        )
        i = 0
        for option in self.preheat_options:
            if option != "cooldown":
                self.labels[option] = self._gtk.Button(
                    label=option, style=f"color{(i % 4) + 1}"
                )
                self.labels[option].connect("clicked", self.set_temperature, option)
                self.labels["preheat_grid"].attach(
                    self.labels[option], (i % 2), int(i / 2), 1, 1
                )
                i += 1
        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.labels["preheat_grid"])
        return scroll

    def delta_adjust(self):
        # Create buttons for increase and decrease
        self.labels["increase"] = self._gtk.Button("increase", None, "color1")
        self.labels["increase"].connect("clicked", self.change_target_temp_incremental, "+")
        self.labels["decrease"] = self._gtk.Button("decrease", None, "color3")
        self.labels["decrease"].connect("clicked", self.change_target_temp_incremental, "-")

        # Create buttons for temperature deltas
        for i in self.tempdeltas:
            self.labels[f"deg{i}"] = self._gtk.Button(label=i)
            self.labels[f"deg{i}"].connect("clicked", self.change_temp_delta, i)
            ctx = self.labels[f"deg{i}"].get_style_context()
            ctx.add_class("horizontal_togglebuttons")
            if i == self.tempdelta:
                ctx.add_class("horizontal_togglebuttons_active")

        # Create grid for temperature deltas
        tempgrid = Gtk.Grid()
        for j, i in enumerate(self.tempdeltas):
            tempgrid.attach(self.labels[f"deg{i}"], j, 0, 1, 1)

        # Create grid for decrease button, increase button, temperature labels, and grid
        deltagrid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        if self._screen.vertical_mode:
            deltagrid.attach(self.labels["decrease"], 0, 1, 1, 1)
            deltagrid.attach(self.labels["increase"], 1, 1, 1, 1)
            deltagrid.attach(tempgrid, 0, 2, 2, 1)
        else:
            deltagrid.attach(self.labels["decrease"], 0, 1, 1, 3)
            deltagrid.attach(self.labels["increase"], 1, 1, 1, 3)
            deltagrid.attach(Gtk.Label(_("Temperature") + " (°C)"), 0, 4, 2, 1)
            deltagrid.attach(tempgrid, 0, 5, 2, 2)
        return deltagrid

    def change_temp_delta(self, widget, tempdelta):
        logging.info(f"### tempdelta {tempdelta}")
        self.labels[f"deg{self.tempdelta}"].get_style_context().remove_class(
            "horizontal_togglebuttons_active"
        )
        self.labels[f"deg{tempdelta}"].get_style_context().add_class(
            "horizontal_togglebuttons_active"
        )
        self.tempdelta = tempdelta

    def change_target_temp_incremental(self, widget, direction):

        if len(self.active_heaters) == 0:
            self._screen.show_popup_message(_("Nothing selected"))
        else:
            for heater in self.active_heaters:
                target = self._printer.get_stat(heater, "target")
                name = heater.split()[1] if len(heater.split()) > 1 else heater
                if direction == "+":
                    target += int(self.tempdelta)
                    max_temp = int(
                        float(self._printer.get_config_section(heater)["max_temp"])
                    )
                    if target > max_temp:
                        target = max_temp
                        self._screen.show_popup_message(
                            _("Can't set above the maximum:") + f" {target}"
                        )

                else:
                    target -= int(self.tempdelta)
                    target = max(target, 0)
                if heater.startswith("extruder"):
                    self._screen._ws.klippy.set_tool_temp(
                        self._printer.get_tool_number(heater), target
                    )
                elif heater.startswith("heater_bed"):
                    self._screen._ws.klippy.set_bed_temp(target)
                elif heater.startswith("heater_generic "):
                    self._screen._ws.klippy.set_heater_temp(name, target)
                elif heater.startswith("temperature_fan "):
                    self._screen._ws.klippy.set_temp_fan_temp(name, target)
                else:
                    logging.info(f"Unknown heater: {heater}")
                    self._screen.show_popup_message(_("Unknown Heater") + " " + heater)
                self._printer.set_stat(heater, {"target": target})
                logging.info(f"Setting {heater} to {target}")

    def update_graph_visibility(self, force_hide=False):
        count = 0
        for device in self.devices:
            visible = self._config.get_config().getboolean(
                f"graph {self._screen.connected_printer}", device, fallback=True
            )
            self.devices[device]["visible"] = visible
            self.labels["da"].set_showing(device, visible)
            if visible:
                count += 1
                self.devices[device]["name_button"].get_style_context().add_class(
                    "graph_label"
                )
            else:
                self.devices[device]["name_button"].get_style_context().remove_class(
                    "graph_label"
                )
        if count > 0 and not force_hide:
            if self.labels["da"] not in self.left_panel:
                self.left_panel.add(self.labels["da"])
            self.labels["da"].queue_draw()
            self.labels["da"].show()
            if self.graph_update is None:
                # This has a high impact on load
                self.graph_update = GLib.timeout_add_seconds(5, self.update_graph)
        elif self.labels["da"] in self.left_panel:
            self.left_panel.remove(self.labels["da"])
            if self.graph_update is not None:
                GLib.source_remove(self.graph_update)
                self.graph_update = None

    def activate(self):
        if not self._printer.tempstore:
            self._screen.init_tempstore()
        self.update_graph_visibility()
        self.set_selection()

    def set_extra(self, extra=None, **kwargs):
        self.extra_selection = extra

    def set_selection(self):
        selection = []
        if self.extra_selection:
            selection.append(self.extra_selection)
        elif self._printer.state not in ("printing", "paused"):
            selection.extend(self._printer.get_temp_devices())
        elif 'toolhead' in self._printer.data and 'extruder' in self._printer.data['toolhead']:
            current_extruder = self._printer.data['toolhead']['extruder']
            selection.append(current_extruder)

        for heater in self.active_heaters:
            if heater not in selection:
                self.select_heater(None, heater)
        for heater in selection:
            if heater.startswith("temperature_sensor "):
                continue
            name = heater.split()[1] if len(heater.split()) > 1 else heater
            if heater not in self.active_heaters:
                self.select_heater(None, heater)
        self.extra_selection = None

    def deactivate(self):
        if self.graph_update is not None:
            GLib.source_remove(self.graph_update)
            self.graph_update = None
        if self.active_heater is not None:
            self.hide_numpad()

    def select_heater(self, widget, device):
        if (
            self.active_heater is None
            and device in self.devices
            and self._printer.device_has_target(device)
        ):
            if device in self.active_heaters:
                self.active_heaters.pop(self.active_heaters.index(device))
                self.devices[device]["name_button"].get_style_context().remove_class(
                    "button_active"
                )
                logging.info(f"Deselecting {device}")
                return
            self.active_heaters.append(device)
            self.devices[device]["name_button"].get_style_context().add_class(
                "button_active"
            )
            logging.info(f"Selecting {device}")
        return

    def set_temperature(self, widget, setting):
        if len(self.active_heaters) == 0:
            self._screen.show_popup_message(_("Nothing selected"))
        else:
            for heater in self.active_heaters:
                target = None
                max_temp = float(self._printer.get_config_section(heater)["max_temp"])
                name = heater.split()[1] if len(heater.split()) > 1 else heater
                with suppress(KeyError):
                    for i in self.preheat_options[setting]:
                        logging.info(f"{self.preheat_options[setting]}")
                        if i == name:
                            # Assign the specific target if available
                            target = self.preheat_options[setting][name]
                            logging.info(f"name match {name}")
                        elif i == heater:
                            target = self.preheat_options[setting][heater]
                            logging.info(f"heater match {heater}")
                if (
                    target is None
                    and setting == "cooldown"
                    and not heater.startswith("temperature_fan ")
                ):
                    target = 0
                if heater.startswith("extruder"):
                    if self.validate(heater, target, max_temp):
                        self._screen._ws.klippy.set_tool_temp(
                            self._printer.get_tool_number(heater), target
                        )
                elif heater.startswith("heater_bed"):
                    if target is None:
                        with suppress(KeyError):
                            target = self.preheat_options[setting]["bed"]
                    if self.validate(heater, target, max_temp):
                        self._screen._ws.klippy.set_bed_temp(target)
                elif heater.startswith("heater_generic "):
                    if target is None:
                        with suppress(KeyError):
                            target = self.preheat_options[setting]["heater_generic"]
                    if self.validate(heater, target, max_temp):
                        self._screen._ws.klippy.set_heater_temp(name, target)
                elif heater.startswith("temperature_fan "):
                    if target is None:
                        with suppress(KeyError):
                            target = self.preheat_options[setting]["temperature_fan"]
                    if self.validate(heater, target, max_temp):
                        self._screen._ws.klippy.set_temp_fan_temp(name, target)
            # This small delay is needed to properly update the target if the user configured something above
            # and then changed the target again using preheat gcode
            GLib.timeout_add(250, self.preheat_gcode, widget, setting)

    def validate(self, heater, target=None, max_temp=None):
        if target is not None and max_temp is not None:
            if 0 <= target <= max_temp:
                return True
            elif target > max_temp:
                self._screen.show_popup_message(
                    _("Can't set above the maximum:") + f" {max_temp}"
                )
                return False
        logging.debug(f"Invalid {heater} Target:{target}/{max_temp}")
        return False

    def preheat_gcode(self, widget, setting):
        with suppress(KeyError):
            script = {"script": self.preheat_options[setting]["gcode"]}
            self._screen._send_action(widget, "printer.gcode.script", script)
        return False

    def add_device(self, device):

        logging.info(f"Adding device: {device}")

        temperature = self._printer.get_stat(device, "temperature")
        if temperature is None:
            return False

        devname = device.split()[1] if len(device.split()) > 1 else device
        # Support for hiding devices by name
        if devname.startswith("_"):
            return False
        if devname.lower() in self.hidden_sensors:
            return False

        if device.startswith("extruder"):
            if self._printer.extrudercount > 1:
                image = f"extruder-{device[8:]}" if device[8:] else "extruder-0"
            else:
                image = "extruder"
            class_name = f"graph_label_{device}"
            dev_type = "extruder"
        elif device == "heater_bed":
            image = "bed"
            devname = "Heater Bed"
            class_name = "graph_label_heater_bed"
            dev_type = "bed"
        elif device.startswith("heater_generic"):
            self.h += 1
            image = "heater"
            class_name = f"graph_label_sensor_{self.h}"
            dev_type = "sensor"
        elif device.startswith("temperature_fan"):
            self.f += 1
            image = "fan"
            class_name = f"graph_label_fan_{self.f}"
            dev_type = "fan"
        elif self._config.get_main_config().getboolean("only_heaters", False):
            return False
        else:
            self.h += 1
            image = "heat-up"
            class_name = f"graph_label_sensor_{self.h}"
            dev_type = "sensor"

        rgb = self._gtk.get_temp_color(dev_type)

        name = self._gtk.Button(
            image, self.prettify(devname), None, self.bts, Gtk.PositionType.LEFT, 1
        )
        name.set_alignment(0, 0.5)
        name.get_style_context().add_class(class_name)
        visible = self._config.get_config().getboolean(
            f"graph {self._screen.connected_printer}", device, fallback=True
        )
        if visible:
            name.get_style_context().add_class("graph_label")

        self.labels["da"].add_object(device, "temperatures", rgb, False, False)
        temp = self._gtk.Button(label="", lines=1)
        find_widget(temp, Gtk.Label).set_ellipsize(False)

        if self._printer.device_has_target(device):
            temp.connect("clicked", self.show_numpad, device)
            self.labels["da"].add_object(device, "targets", rgb, False, True)
            name.connect("button-press-event", self.name_pressed, device)
            self.long_press[device] = Gtk.GestureLongPress.new(name)
            self.long_press[device].connect(
                "pressed", self.name_long_press, name, device
            )
            self.long_press[device].connect(
                "cancelled", self.name_long_press_cancelled, name, device
            )
        else:
            name.connect("clicked", self.toggle_visibility, device)
        if self._show_heater_power and self._printer.device_has_power(device):
            self.labels["da"].add_object(device, "powers", rgb, True, False)
        self.labels["da"].set_showing(device, visible)

        self.devices[device] = {
            "class": class_name,
            "name_button": name,
            "temp": temp,
            "visible": visible,
        }

        devices = sorted(self.devices)
        pos = devices.index(device) + 1

        self.labels["devices"].insert_row(pos)
        self.labels["devices"].attach(name, 0, pos, 1, 1)
        self.labels["devices"].attach(temp, 1, pos, 1, 1)
        self.labels["devices"].show_all()
        return True

    def name_pressed(self, widget, event, device):
        self.popover_device = device
        if event.button == 3:
            self.popover_popup(widget, device)

    def name_long_press_cancelled(self, gesture_long_press, widget, device):
        if self.active_heater:
            self.show_numpad(widget, device)
        else:
            self.select_heater(widget, device)

    def name_long_press(self, gesture_long_press, x, y, widget, device):
        self.popover_device = device
        self.popover_popup(widget, device)

    def toggle_visibility(self, widget, device=None):
        if device is None:
            device = self.popover_device
        self.devices[device]["visible"] ^= True
        logging.info(f"Graph show {self.devices[device]['visible']}: {device}")

        section = f"graph {self._screen.connected_printer}"
        if section not in self._config.get_config().sections():
            self._config.get_config().add_section(section)
        self._config.set(section, f"{device}", f"{self.devices[device]['visible']}")
        self._config.save_user_config_options()

        self.update_graph_visibility()
        if self._printer.device_has_target(device):
            self.popover_populate_menu()
            self.popover.show_all()

    def change_target_temp(self, temp):
        name = (
            self.active_heater.split()[1]
            if len(self.active_heater.split()) > 1
            else self.active_heater
        )
        temp = self.verify_max_temp(temp)
        if temp is False:
            return

        if self.active_heater.startswith("extruder"):
            self._screen._ws.klippy.set_tool_temp(
                self._printer.get_tool_number(self.active_heater), temp
            )
        elif self.active_heater == "heater_bed":
            self._screen._ws.klippy.set_bed_temp(temp)
        elif self.active_heater.startswith("heater_generic "):
            self._screen._ws.klippy.set_heater_temp(name, temp)
        elif self.active_heater.startswith("temperature_fan "):
            self._screen._ws.klippy.set_temp_fan_temp(name, temp)
        else:
            logging.info(f"Unknown heater: {self.active_heater}")
            self._screen.show_popup_message(
                _("Unknown Heater") + " " + self.active_heater
            )
        self._printer.set_stat(name, {"target": temp})
        if self.numpad_visible:
            self.hide_numpad()

    def verify_max_temp(self, temp):
        temp = int(temp)
        max_temp = int(
            float(self._printer.get_config_section(self.active_heater)["max_temp"])
        )
        logging.debug(f"{temp}/{max_temp}")
        if temp > max_temp:
            self._screen.show_popup_message(
                _("Can't set above the maximum:") + f" {max_temp}"
            )
            return False
        return max(temp, 0)

    def pid_calibrate(self, temp):
        if temp <= 0:
            return
        heater = self.active_heater.split(' ', maxsplit=1)[-1]
        if self.verify_max_temp(temp):
            script = {
                "script": f"PID_CALIBRATE HEATER={heater} TARGET={temp}"
            }
            self._screen._confirm_send_action(
                None,
                _("Initiate a PID calibration for:")
                + f" {heater} @ {temp} ºC"
                + "\n\n"
                + _("It may take more than 5 minutes depending on the heater power."),
                "printer.gcode.script",
                script,
            )

    def create_left_panel(self):

        self.labels["devices"] = Gtk.Grid(vexpand=False)
        self.labels["devices"].get_style_context().add_class("heater-grid")

        name = Gtk.Label()
        temp = Gtk.Label(_("Temp (°C)"))

        self.labels["devices"].attach(name, 0, 0, 1, 1)
        self.labels["devices"].attach(temp, 1, 0, 1, 1)

        self.labels["da"] = HeaterGraph(
            self._screen, self._printer, self._gtk.font_size
        )

        scroll = self._gtk.ScrolledWindow(steppers=False)
        scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scroll.get_style_context().add_class("heater-list")
        scroll.add(self.labels["devices"])

        self.left_panel = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.left_panel.add(scroll)

        self.popover_buttons = {
            "set_temp": self._gtk.Button(label=_("Set Temp")),
            "graph_show": self._gtk.Button(label=_("Show")),
        }
        self.popover_buttons["set_temp"].connect("clicked", self.show_numpad)
        self.popover_buttons["set_temp"].set_no_show_all(True)
        self.popover_buttons["graph_show"].connect("clicked", self.toggle_visibility)

        pobox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        pobox.pack_start(self.popover_buttons["graph_show"], True, True, 5)
        pobox.pack_start(self.popover_buttons["set_temp"], True, True, 5)
        self.popover.add(pobox)
        self.popover.connect("closed", self.popover_closed)

        for d in self._printer.get_temp_devices():
            self.add_device(d)

        return self.left_panel

    def hide_numpad(self, widget=None):
        self.devices[self.active_heater][
            "name_button"
        ].get_style_context().remove_class("button_active")
        self.active_heater = None

        for d in self.active_heaters:
            self.devices[d]["name_button"].get_style_context().add_class(
                "button_active"
            )

        if self._screen.vertical_mode:
            if not self._gtk.ultra_tall:
                self.update_graph_visibility(force_hide=False)
            top = self.grid.get_child_at(0, 0)
            bottom = self.grid.get_child_at(0, 2)
            self.grid.remove(top)
            self.grid.remove(bottom)
            self.grid.attach(top, 0, 0, 1, 3)
            self.grid.attach(self.create_right_panel(), 0, 3, 1, 2)
        else:
            self.grid.remove_column(1)
            self.grid.attach(self.create_right_panel(), 1, 0, 1, 1)
        self.grid.show_all()
        self.numpad_visible = False

    def popover_closed(self, widget):
        self.popover_device = None

    def popover_popup(self, widget, device):
        self.popover_device = device
        self.popover.set_relative_to(widget)
        self.popover_populate_menu()
        self.popover.show_all()

    def popover_populate_menu(self):
        if self.labels["da"].is_showing(self.popover_device):
            self.popover_buttons["graph_show"].set_label(_("Hide"))
        else:
            self.popover_buttons["graph_show"].set_label(_("Show"))
        if self._printer.device_has_target(self.popover_device):
            self.popover_buttons["set_temp"].show()
        else:
            self.popover_buttons["set_temp"].hide()

    def process_update(self, action, data):
        if action != "notify_status_update":
            return
        for x in self._printer.get_temp_devices():
            if x in data:
                if x not in self.devices:
                    self.add_device(x)
                self.update_temp(
                    x,
                    self._printer.get_stat(x, "temperature"),
                    self._printer.get_stat(x, "target"),
                    self._printer.get_stat(x, "power"),
                )

    def show_numpad(self, widget, device=None):
        for d in self.active_heaters:
            self.devices[d]["name_button"].get_style_context().remove_class(
                "button_active"
            )
        self.active_heater = self.popover_device if device is None else device
        self.devices[self.active_heater]["name_button"].get_style_context().add_class(
            "button_active"
        )
        if "keypad" not in self.labels:
            self.labels["keypad"] = Keypad(
                self._screen,
                self.change_target_temp,
                self.hide_numpad,
                error_msg=_("Invalid temperature")
            )
        if (
            self._printer.state not in ("printing", "paused")
            and self._screen.printer.config[self.active_heater]["control"] == "pid"
        ):
            self.labels["keypad"].add_extra_button(
                self.pid_calibrate,
                icon="heat-up",
                label=_('Calibrate') + ' PID',
            )
        self.labels["keypad"].clear()

        if self._screen.vertical_mode:
            if not self._gtk.ultra_tall:
                self.update_graph_visibility(force_hide=True)
            top = self.grid.get_child_at(0, 0)
            bottom = self.grid.get_child_at(0, 3)
            self.grid.remove(top)
            self.grid.remove(bottom)
            self.grid.attach(top, 0, 0, 1, 2)
            self.grid.attach(self.labels["keypad"], 0, 2, 1, 2)
        else:
            self.grid.remove_column(1)
            self.grid.attach(self.labels["keypad"], 1, 0, 1, 1)
        self.grid.show_all()
        self.numpad_visible = True
        self.popover.popdown()

    def update_graph(self):
        self.labels["da"].queue_draw()
        return True



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
from panels.menu import Panel as MenuPanel
from ks_includes.widgets.heatergraph import HeaterGraph
from ks_includes.widgets.keypad import Keypad
from ks_includes.KlippyGtk import find_widget


class Panel(MenuPanel):
    def __init__(self, screen, title, items=None):
        super().__init__(screen, title, items)
        self.left_panel = None
        self.devices = {}
        self.graph_update = None
        self.active_heater = None
        self.h = self.f = 0
        self.main_menu = Gtk.Grid(row_homogeneous=True, column_homogeneous=True, hexpand=True, vexpand=True)
        scroll = self._gtk.ScrolledWindow()
        self.numpad_visible = False

        logging.info("### Making MainMenu")

        stats = self._printer.get_printer_status_data()["printer"]
        if stats["temperature_devices"]["count"] > 0 or stats["extruders"]["count"] > 0:
            self._gtk.reset_temp_color()
        if self._screen.vertical_mode:
            self.main_menu.attach(self.create_left_panel(), 0, 0, 1, 3)
            self.labels['menu'] = self.arrangeMenuItems(items, 3, True)
            scroll.add(self.labels['menu'])
            self.main_menu.attach(scroll, 0, 3, 1, 2)
        else:
            self.main_menu.attach(self.create_left_panel(), 0, 0, 1, 1)
            self.labels['menu'] = self.arrangeMenuItems(items, 2, True)
            scroll.add(self.labels['menu'])
            self.main_menu.attach(scroll, 1, 0, 1, 1)
        self.content.add(self.main_menu)

    def update_graph_visibility(self, force_hide=False):
        if self.left_panel is None:
            logging.info("No left panel")
            return
        count = 0
        for device in self.devices:
            visible = self._config.get_config().getboolean(f"graph {self._screen.connected_printer}",
                                                           device, fallback=True)
            self.devices[device]['visible'] = visible
            self.labels['da'].set_showing(device, visible)
            if visible:
                count += 1
                self.devices[device]['name'].get_style_context().add_class("graph_label")
            else:
                self.devices[device]['name'].get_style_context().remove_class("graph_label")
        if count > 0 and not force_hide:
            if self.labels['da'] not in self.left_panel:
                self.left_panel.add(self.labels['da'])
            self.labels['da'].queue_draw()
            self.labels['da'].show()
            if self.graph_update is None:
                # This has a high impact on load
                self.graph_update = GLib.timeout_add_seconds(5, self.update_graph)
        elif self.labels['da'] in self.left_panel:
            self.left_panel.remove(self.labels['da'])
            if self.graph_update is not None:
                GLib.source_remove(self.graph_update)
                self.graph_update = None
        return False

    def activate(self):
        if not self._printer.tempstore:
            self._screen.init_tempstore()
        self.update_graph_visibility()

    def deactivate(self):
        if self.graph_update is not None:
            GLib.source_remove(self.graph_update)
            self.graph_update = None
        if self.active_heater is not None:
            self.hide_numpad()

    def add_device(self, device):

        logging.info(f"Adding device: {device}")

        temperature = self._printer.get_stat(device, "temperature")
        if temperature is None:
            return False

        devname = device.split()[1] if len(device.split()) > 1 else device
        # Support for hiding devices by name
        if devname.startswith("_"):
            return False
        if devname.lower() in self.hidden_sensors:
            return False

        if device.startswith("extruder"):
            if self._printer.extrudercount > 1:
                image = f"extruder-{device[8:]}" if device[8:] else "extruder-0"
            else:
                image = "extruder"
            class_name = f"graph_label_{device}"
            dev_type = "extruder"
        elif device == "heater_bed":
            image = "bed"
            devname = "Heater Bed"
            class_name = "graph_label_heater_bed"
            dev_type = "bed"
        elif device.startswith("heater_generic"):
            self.h += 1
            image = "heater"
            class_name = f"graph_label_sensor_{self.h}"
            dev_type = "sensor"
        elif device.startswith("temperature_fan"):
            self.f += 1
            image = "fan"
            class_name = f"graph_label_fan_{self.f}"
            dev_type = "fan"
        elif self._config.get_main_config().getboolean("only_heaters", False):
            return False
        else:
            self.h += 1
            image = "heat-up"
            class_name = f"graph_label_sensor_{self.h}"
            dev_type = "sensor"

        rgb = self._gtk.get_temp_color(dev_type)

        can_target = self._printer.device_has_target(device)
        self.labels['da'].add_object(device, "temperatures", rgb, False, False)
        if can_target:
            self.labels['da'].add_object(device, "targets", rgb, False, True)
        if self._show_heater_power and self._printer.device_has_power(device):
            self.labels['da'].add_object(device, "powers", rgb, True, False)

        name = self._gtk.Button(image, self.prettify(devname), None, self.bts, Gtk.PositionType.LEFT, 1)
        name.connect("clicked", self.toggle_visibility, device)
        name.set_alignment(0, .5)
        name.get_style_context().add_class(class_name)
        visible = self._config.get_config().getboolean(f"graph {self._screen.connected_printer}", device, fallback=True)
        if visible:
            name.get_style_context().add_class("graph_label")
        self.labels['da'].set_showing(device, visible)

        temp = self._gtk.Button(label="", lines=1)
        find_widget(temp, Gtk.Label).set_ellipsize(False)
        if can_target:
            temp.connect("clicked", self.show_numpad, device)

        self.devices[device] = {
            "class": class_name,
            "name": name,
            "temp": temp,
            "can_target": can_target,
            "visible": visible
        }

        devices = sorted(self.devices)
        pos = devices.index(device) + 1

        self.labels['devices'].insert_row(pos)
        self.labels['devices'].attach(name, 0, pos, 1, 1)
        self.labels['devices'].attach(temp, 1, pos, 1, 1)
        self.labels['devices'].show_all()
        return True

    def toggle_visibility(self, widget, device):
        self.devices[device]['visible'] ^= True
        logging.info(f"Graph show {self.devices[device]['visible']}: {device}")

        section = f"graph {self._screen.connected_printer}"
        if section not in self._config.get_config().sections():
            self._config.get_config().add_section(section)
        self._config.set(section, f"{device}", f"{self.devices[device]['visible']}")
        self._config.save_user_config_options()

        self.update_graph_visibility()

    def change_target_temp(self, temp):
        name = self.active_heater.split()[1] if len(self.active_heater.split()) > 1 else self.active_heater
        temp = self.verify_max_temp(temp)
        if temp is False:
            return

        if self.active_heater.startswith('extruder'):
            self._screen._ws.klippy.set_tool_temp(self._printer.get_tool_number(self.active_heater), temp)
        elif self.active_heater == "heater_bed":
            self._screen._ws.klippy.set_bed_temp(temp)
        elif self.active_heater.startswith('heater_generic '):
            self._screen._ws.klippy.set_heater_temp(name, temp)
        elif self.active_heater.startswith('temperature_fan '):
            self._screen._ws.klippy.set_temp_fan_temp(name, temp)
        else:
            logging.info(f"Unknown heater: {self.active_heater}")
            self._screen.show_popup_message(_("Unknown Heater") + " " + self.active_heater)
        self._printer.set_stat(name, {"target": temp})
        if self.numpad_visible:
            self.hide_numpad()

    def verify_max_temp(self, temp):
        temp = int(temp)
        max_temp = int(float(self._printer.get_config_section(self.active_heater)['max_temp']))
        logging.debug(f"{temp}/{max_temp}")
        if temp > max_temp:
            self._screen.show_popup_message(_("Can't set above the maximum:") + f' {max_temp}')
            return False
        return max(temp, 0)

    def pid_calibrate(self, temp):
        if temp <= 0:
            return
        heater = self.active_heater.split(' ', maxsplit=1)[-1]
        if self.verify_max_temp(temp):
            script = {"script": f"PID_CALIBRATE HEATER={heater} TARGET={temp}"}
            self._screen._confirm_send_action(
                None,
                _("Initiate a PID calibration for:")
                + f" {heater} @ {temp} ºC"
                + "\n\n"
                + _("It may take more than 5 minutes depending on the heater power."),
                "printer.gcode.script",
                script
            )

    def create_left_panel(self):

        self.labels['devices'] = Gtk.Grid(vexpand=False)
        self.labels['devices'].get_style_context().add_class('heater-grid')

        name = Gtk.Label()
        temp = Gtk.Label(label=_("Temp (°C)"))

        self.labels['devices'].attach(name, 0, 0, 1, 1)
        self.labels['devices'].attach(temp, 1, 0, 1, 1)

        self.labels['da'] = HeaterGraph(self._screen, self._printer, self._gtk.font_size)

        scroll = self._gtk.ScrolledWindow(steppers=False)
        scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scroll.get_style_context().add_class('heater-list')
        scroll.add(self.labels['devices'])

        self.left_panel = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.left_panel.add(scroll)

        for d in self._printer.get_temp_devices():
            self.add_device(d)

        return self.left_panel

    def hide_numpad(self, widget=None):
        self.devices[self.active_heater]['name'].get_style_context().remove_class("button_active")
        self.active_heater = None

        if self._screen.vertical_mode:
            if not self._gtk.ultra_tall:
                self.update_graph_visibility(force_hide=False)
            top = self.main_menu.get_child_at(0, 0)
            bottom = self.main_menu.get_child_at(0, 2)
            self.main_menu.remove(top)
            self.main_menu.remove(bottom)
            self.main_menu.attach(top, 0, 0, 1, 3)
            self.main_menu.attach(self.labels["menu"], 0, 3, 1, 2)
        else:
            self.main_menu.remove_column(1)
            self.main_menu.attach(self.labels["menu"], 1, 0, 1, 1)
        self.main_menu.show_all()
        self.numpad_visible = False
        self._screen.base_panel.set_control_sensitive(False, control='back')

    def process_update(self, action, data):
        if action != "notify_status_update":
            return
        for x in self._printer.get_temp_devices():
            if x in data:
                if x not in self.devices:
                    self.add_device(x)
                self.update_temp(
                    x,
                    self._printer.get_stat(x, "temperature"),
                    self._printer.get_stat(x, "target"),
                    self._printer.get_stat(x, "power"),
                )

    def show_numpad(self, widget, device):

        if self.active_heater is not None:
            self.devices[self.active_heater]['name'].get_style_context().remove_class("button_active")
        self.active_heater = device
        self.devices[self.active_heater]['name'].get_style_context().add_class("button_active")

        if "keypad" not in self.labels:
            self.labels["keypad"] = Keypad(
                self._screen,
                self.change_target_temp,
                self.hide_numpad,
                error_msg=_("Invalid temperature")
            )
        if (
            self._printer.state not in ("printing", "paused")
            and self._screen.printer.config[self.active_heater]["control"] == "pid"
        ):
            self.labels["keypad"].add_extra_button(
                self.pid_calibrate,
                icon="heat-up",
                label=_('Calibrate') + ' PID',
            )
        self.labels["keypad"].clear()

        if self._screen.vertical_mode:
            if not self._gtk.ultra_tall:
                self.update_graph_visibility(force_hide=True)
            top = self.main_menu.get_child_at(0, 0)
            bottom = self.main_menu.get_child_at(0, 3)
            self.main_menu.remove(top)
            self.main_menu.remove(bottom)
            self.main_menu.attach(top, 0, 0, 1, 2)
            self.main_menu.attach(self.labels["keypad"], 0, 2, 1, 2)
        else:
            self.main_menu.remove_column(1)
            self.main_menu.attach(self.labels["keypad"], 1, 0, 1, 1)
        self.main_menu.show_all()
        self.numpad_visible = True
        self._screen.base_panel.set_control_sensitive(True, control='back')

    def update_graph(self):
        self.labels['da'].queue_draw()
        return True

    def back(self):
        if self.numpad_visible:
            self.hide_numpad()
            return True
        return False



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("Power")
        super().__init__(screen, title)
        self.devices = {}

        # Create a grid for all devices
        self.labels['devices'] = Gtk.Grid(valign=Gtk.Align.CENTER)

        self.load_power_devices()

        # Create a scroll window for the power devices
        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.labels['devices'])

        self.content.add(scroll)

    def activate(self):
        devices = self._printer.get_power_devices()
        for x in devices:
            self.devices[x]['switch'].disconnect_by_func(self.on_switch)
            self.devices[x]['switch'].set_active(self._printer.get_power_device_status(x) == "on")

            self.devices[x]['switch'].connect("notify::active", self.on_switch, x)

    def add_device(self, device):
        name = Gtk.Label(
            hexpand=True, vexpand=True, halign=Gtk.Align.START, valign=Gtk.Align.CENTER,
            wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        name.set_markup(f"<big><b>{device}</b></big>")
        switch = Gtk.Switch(hexpand=False, active=(self._printer.get_power_device_status(device) == "on"),
                            width_request=round(self._gtk.font_size * 7),
                            height_request=round(self._gtk.font_size * 3.5))
        switch.connect("notify::active", self.on_switch, device)

        labels = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        labels.add(name)

        dev = Gtk.Box(
            spacing=5, hexpand=True, vexpand=False, valign=Gtk.Align.CENTER)
        dev.add(labels)
        dev.add(switch)

        self.devices[device] = {
            "row": dev,
            "switch": switch
        }

        devices = sorted(self.devices)
        pos = devices.index(device)

        self.labels['devices'].insert_row(pos)
        self.labels['devices'].attach(self.devices[device]['row'], 0, pos, 1, 1)
        self.labels['devices'].show_all()

    def load_power_devices(self):
        devices = self._printer.get_power_devices()
        for x in devices:
            self.add_device(x)

    def on_switch(self, switch, gparam, device):
        logging.debug(f"Power toggled {device}")
        if switch.get_active():
            self._screen._ws.klippy.power_device_on(device)
        else:
            self._screen._ws.klippy.power_device_off(device)

    def process_update(self, action, data):
        if action != "notify_power_changed":
            return

        if data['device'] not in self.devices:
            return
        device = data['device']
        self.devices[device]['switch'].disconnect_by_func(self.on_switch)
        self.devices[device]['switch'].set_active(data['status'] == "on")
        self.devices[device]['switch'].connect("notify::active", self.on_switch, device)



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib, Pango
from ks_includes.screen_panel import ScreenPanel


CHANGEABLE_FANS = ["fan", "fan_generic"]


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("Fan")
        super().__init__(screen, title)
        self.fan_speed = {}
        self.devices = {}
        # Create a grid for all devices
        self.labels['devices'] = Gtk.Grid(valign=Gtk.Align.CENTER)

        self.load_fans()

        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.labels['devices'])

        self.content.add(scroll)

    def process_update(self, action, data):
        if action != "notify_status_update":
            return

        for fan in self.devices:
            if fan in data and "speed" in data[fan]:
                self.update_fan_speed(None, fan, self._printer.get_fan_speed(fan))

    def update_fan_speed(self, widget, fan, speed):
        if fan not in self.devices:
            return

        if self.devices[fan]['changeable'] is True:
            if self.devices[fan]['scale'].has_grab():
                return
            self.devices[fan]["speed"] = round(float(speed) * 100)
            self.devices[fan]['scale'].disconnect_by_func(self.set_fan_speed)
            self.devices[fan]['scale'].set_value(self.devices[fan]["speed"])
            self.devices[fan]['scale'].connect("button-release-event", self.set_fan_speed, fan)
        else:
            self.devices[fan]["speed"] = float(speed)
            self.devices[fan]['scale'].set_fraction(self.devices[fan]["speed"])
        if widget is not None:
            self.set_fan_speed(None, None, fan)

    def add_fan(self, fan):

        logging.info(f"Adding fan: {fan}")
        changeable = any(fan.startswith(x) or fan == x for x in CHANGEABLE_FANS)
        name = Gtk.Label(halign=Gtk.Align.START, valign=Gtk.Align.CENTER, hexpand=True, vexpand=True,
                         wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        fan_name = _("Part Fan") if fan == "fan" else fan.split()[1]
        name.set_markup(f"\n<big><b>{fan_name}</b></big>\n")

        fan_col = Gtk.Box(spacing=5)
        stop_btn = self._gtk.Button("cancel", None, "color1")
        stop_btn.set_hexpand(False)
        stop_btn.connect("clicked", self.update_fan_speed, fan, 0)
        max_btn = self._gtk.Button("fan-on", _("Max"), "color2")
        max_btn.set_hexpand(False)
        max_btn.connect("clicked", self.update_fan_speed, fan, 100)

        speed = float(self._printer.get_fan_speed(fan))
        if changeable:
            speed = round(speed * 100)
            scale = Gtk.Scale.new_with_range(Gtk.Orientation.HORIZONTAL, min=0, max=100, step=1)
            scale.set_value(speed)
            scale.set_digits(0)
            scale.set_hexpand(True)
            scale.set_has_origin(True)
            scale.get_style_context().add_class("fan_slider")
            scale.connect("button-release-event", self.set_fan_speed, fan)
            fan_col.add(stop_btn)
            fan_col.add(scale)
            fan_col.add(max_btn)
        else:
            scale = Gtk.ProgressBar(hexpand=True, show_text=True, fraction=speed)
            fan_col.pack_start(scale, True, True, 10)

        fan_row = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        fan_row.add(name)
        fan_row.add(fan_col)

        self.devices[fan] = {
            "changeable": changeable,
            "scale": scale,
            "speed": speed,
        }

        devices = sorted(self.devices)
        if fan == "fan":
            pos = 0
        elif "fan" in devices:
            devices.pop(devices.index("fan"))
            pos = devices.index(fan) + 1
        else:
            pos = devices.index(fan)

        self.labels['devices'].insert_row(pos)
        self.labels['devices'].attach(fan_row, 0, pos, 1, 1)
        self.labels['devices'].show_all()

    def load_fans(self):
        fans = self._printer.get_fans()
        for fan in fans:
            # Support for hiding devices by name
            name = fan.split()[1] if len(fan.split()) > 1 else fan
            if name.startswith("_"):
                continue
            self.add_fan(fan)

    def set_fan_speed(self, widget, event, fan):
        value = self.devices[fan]['scale'].get_value()

        if fan == "fan":
            self._screen._ws.klippy.gcode_script(f"M106 S{value * 2.55:.0f}")
        else:
            self._screen._ws.klippy.gcode_script(f"SET_FAN_SPEED FAN={fan.split()[1]} SPEED={float(value) / 100}")
        # Check the speed in case it wasn't applied
        GLib.timeout_add_seconds(1, self.check_fan_speed, fan)

    def check_fan_speed(self, fan):
        self.update_fan_speed(None, fan, self._printer.get_fan_speed(fan))
        return False



import logging
import os

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib, Pango
from ks_includes.screen_panel import ScreenPanel
from ks_includes.sdbus_nm import SdbusNm
from datetime import datetime


class Panel(ScreenPanel):

    def __init__(self, screen, title):
        title = title or _("Network")
        super().__init__(screen, title)
        self.last_drop_time = datetime.now()
        self.show_add = False
        try:
            self.sdbus_nm = SdbusNm(self.popup_callback)
        except Exception as e:
            logging.exception("Failed to initialize")
            self.sdbus_nm = None
            self.error_box = Gtk.Box(
                orientation=Gtk.Orientation.VERTICAL,
                hexpand=True,
                vexpand=True
            )
            message = (
                _("Failed to initialize") + "\n"
                + "This panel needs NetworkManager installed into the system\n"
                + "And the appropriate permissions, without them it will not function.\n"
                + f"\n{e}\n"
            )
            self.error_box.add(
                Gtk.Label(
                    label=message,
                    wrap=True,
                    wrap_mode=Pango.WrapMode.WORD_CHAR,
                )
            )
            self.error_box.set_valign(Gtk.Align.CENTER)
            self.content.add(self.error_box)
            self._screen.panels_reinit.append(self._screen._cur_panels[-1])
            return
        self.update_timeout = None
        self.network_list = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, hexpand=True, vexpand=True)
        self.network_rows = {}
        self.networks = {}
        self.wifi_signal_icons = {
            'excellent': self._gtk.PixbufFromIcon('wifi_excellent'),
            'good': self._gtk.PixbufFromIcon('wifi_good'),
            'fair': self._gtk.PixbufFromIcon('wifi_fair'),
            'weak': self._gtk.PixbufFromIcon('wifi_weak'),
        }

        self.network_interfaces = self.sdbus_nm.get_interfaces()
        logging.info(f"Network interfaces: {self.network_interfaces}")

        self.wireless_interfaces = [iface.interface for iface in self.sdbus_nm.get_wireless_interfaces()]
        logging.info(f"Wireless interfaces: {self.wireless_interfaces}")

        self.interface = self.sdbus_nm.get_primary_interface()
        logging.info(f"Primary interface: {self.interface}")

        iface_label = Gtk.Label(_("Interface") + ": ")
        if len(self.wireless_interfaces) <= 1:
            self.labels['interface'] = Gtk.Label(hexpand=True, halign=Gtk.Align.START)
            if self.interface is not None:
                self.labels['interface'].set_text(f'{self.interface}')
        else:
            self.labels['interface'] = Gtk.ComboBoxText()
            self.labels['interface'].connect("changed", self.iface_changed)
            for iface in self.wireless_interfaces:
                self.labels['interface'].append(iface, iface)
            if self.interface is not None:
                self.labels['interface'].set_active_id(self.interface)

        self.labels['ip'] = Gtk.Label(hexpand=True)
        if self.interface is not None:
            self.labels['ip'].set_text(f"IP: {self.sdbus_nm.get_ip_address()}")

        self.reload_button = self._gtk.Button("refresh", None, "color1", self.bts)
        self.reload_button.set_no_show_all(True)
        self.reload_button.show()
        self.reload_button.connect("clicked", self.reload_networks)
        self.reload_button.set_hexpand(False)

        self.wifi_toggle_switch = Gtk.Switch(
            width_request=round(self._gtk.font_size * 2),
            height_request=round(self._gtk.font_size),
            active=self.sdbus_nm.is_wifi_enabled()
        )
        self.wifi_toggle_switch.connect("notify::active", self.toggle_wifi)
        wifi_toggle = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        wifi_toggle.add(self.wifi_toggle_switch)

        sbox = Gtk.Box(hexpand=True, vexpand=False)
        if (self._screen.width > 400):
            sbox.add(iface_label)
        sbox.add(self.labels['interface'])
        sbox.add(self.labels['ip'])
        sbox.add(self.reload_button)
        sbox.add(wifi_toggle)

        scroll = self._gtk.ScrolledWindow()
        self.labels['main_box'] = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, vexpand=True)

        if self.sdbus_nm.wifi:
            self.labels['main_box'].pack_start(sbox, False, False, 5)
            GLib.idle_add(self.load_networks)
            scroll.add(self.network_list)
            self.sdbus_nm.set_connection_monitoring(True)
            self.conn_status = GLib.timeout_add_seconds(1, self.sdbus_nm.monitor_connection_status)
        else:
            self._screen.show_popup_message(_("No wireless interface has been found"), level=2)
            self.labels['networkinfo'] = Gtk.Label()
            scroll.add(self.labels['networkinfo'])
            self.update_single_network_info()

        self.labels['main_box'].pack_start(scroll, True, True, 0)
        self.content.add(self.labels['main_box'])

    def popup_callback(self, msg, level=3):
        self._screen.show_popup_message(msg, level)

    def load_networks(self):
        for net in self.sdbus_nm.get_networks():
            if 'BSSID' not in net:
                continue
            self.add_network(net['BSSID'])
        GLib.timeout_add_seconds(10, self._gtk.Button_busy, self.reload_button, False)
        self.content.show_all()
        return False

    def add_network(self, bssid):
        if bssid in self.network_rows:
            return

        net = next(net for net in self.sdbus_nm.get_networks() if bssid == net['BSSID'])
        ssid = net['SSID']

        connect = self._gtk.Button("load", None, "color3", self.bts)
        connect.connect("clicked", self.connect_network, ssid)
        connect.set_hexpand(False)
        connect.set_halign(Gtk.Align.END)

        delete = self._gtk.Button("delete", None, "color3", self.bts)
        delete.connect("clicked", self.remove_confirm_dialog, ssid, bssid)
        delete.set_hexpand(False)
        delete.set_halign(Gtk.Align.END)

        buttons = Gtk.Box(spacing=5)

        name = Gtk.Label(hexpand=True, halign=Gtk.Align.START, wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        if bssid == self.sdbus_nm.get_connected_bssid():
            ssid += ' (' + _("Connected") + ')'
            name.set_markup(f"<big><b>{ssid}</b></big>")
        else:
            name.set_markup(f"<b>{ssid}</b>")
        if net['known']:
            buttons.add(delete)
        buttons.add(connect)

        info = Gtk.Label(halign=Gtk.Align.START)
        labels = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, vexpand=True,
                         halign=Gtk.Align.START, valign=Gtk.Align.CENTER)
        labels.add(name)
        labels.add(info)
        icon = self._gtk.Image()

        self.network_rows[bssid] = Gtk.Box(spacing=5, hexpand=True, vexpand=False)
        self.network_rows[bssid].get_style_context().add_class("frame-item")
        self.network_rows[bssid].add(icon)
        self.network_rows[bssid].add(labels)
        self.network_rows[bssid].add(buttons)

        self.networks[bssid] = {
            "connect": connect,
            "delete": delete,
            "icon": icon,
            "info": info,
            "name": name,
            "row": self.network_rows[bssid],
        }

        self.network_list.add(self.network_rows[bssid])

    def iface_changed(self, combo):
        selected_iface = combo.get_active_id()
        logging.info(f"Selected interace: {selected_iface}")
        self.interace = selected_iface
        self.sdbus_nm.set_wlan_device(selected_iface)

    def remove_confirm_dialog(self, widget, ssid, bssid):

        label = Gtk.Label(wrap=True, vexpand=True)
        label.set_markup(_("Do you want to forget or disconnect %s?") % ssid)
        buttons = [
            {"name": _("Forget"), "response": Gtk.ResponseType.OK, "style": 'dialog-warning'},
            {"name": _("Cancel"), "response": Gtk.ResponseType.CANCEL, "style": 'dialog-error'},
        ]
        if bssid == self.sdbus_nm.get_connected_bssid():
            buttons.insert(0, {"name": _("Disconnect"), "response": Gtk.ResponseType.APPLY, "style": 'dialog-info'})
        self._gtk.Dialog(_("Remove network"), buttons, label, self.confirm_removal, ssid)

    def confirm_removal(self, dialog, response_id, ssid):
        self._gtk.remove_dialog(dialog)
        if response_id == Gtk.ResponseType.CANCEL:
            return
        bssid = self.sdbus_nm.get_bssid_from_ssid(ssid)
        self.remove_network_from_list(bssid)
        if response_id == Gtk.ResponseType.OK:
            logging.info(f"Deleting {ssid}")
            self.sdbus_nm.delete_network(ssid)
        if response_id == Gtk.ResponseType.APPLY:
            logging.info(f"Disconnecting {ssid}")
            self.sdbus_nm.disconnect_network()

    def add_new_network(self, widget, ssid):
        self._screen.remove_keyboard()
        psk = self.labels['network_psk'].get_text()
        identity = self.labels['network_identity'].get_text()
        eap_method = self.get_dropdown_value(self.labels['network_eap_method'])
        phase2 = self.get_dropdown_value(self.labels['network_phase2'])
        logging.debug(f"{phase2=}")
        logging.debug(f"{eap_method=}")
        result = self.sdbus_nm.add_network(ssid, psk, eap_method, identity, phase2)
        if "error" in result:
            self._screen.show_popup_message(result["message"])
            if result["error"] == "psk_invalid":
                return
        else:
            self.connect_network(widget, ssid, showadd=False)
        self.close_add_network()

    def get_dropdown_value(self, dropdown, default=None):
        tree_iter = dropdown.get_active_iter()
        model = dropdown.get_model()
        result = model[tree_iter][0]
        return result if result != "disabled" else None

    def back(self):
        if self.show_add:
            self.close_add_network()
            return True
        return False

    def close_add_network(self):
        if not self.show_add:
            return

        for child in self.content.get_children():
            self.content.remove(child)
        self.content.add(self.labels['main_box'])
        self.content.show()
        for i in ['add_network', 'network_psk', 'network_identity']:
            if i in self.labels:
                del self.labels[i]
        self.show_add = False

    def connect_network(self, widget, ssid, showadd=True):
        self.deactivate()
        if showadd and not self.sdbus_nm.is_known(ssid):
            sec_type = self.sdbus_nm.get_security_type(ssid)
            if sec_type == "Open" or "OWE" in sec_type:
                logging.debug("Network is Open do not show psk")
                result = self.sdbus_nm.add_network(ssid, '')
                if "error" in result:
                    self._screen.show_popup_message(result["message"])
            else:
                self.show_add_network(widget, ssid)
            self.activate()
            return
        bssid = self.sdbus_nm.get_bssid_from_ssid(ssid)
        if bssid and bssid in self.network_rows:
            self.remove_network_from_list(bssid)
        self.sdbus_nm.connect(ssid)
        self.reload_networks()

    def remove_network_from_list(self, bssid):
        if bssid not in self.network_rows:
            logging.error(f"{bssid} not in rows")
            return
        self.network_list.remove(self.network_rows[bssid])
        del self.network_rows[bssid]
        del self.networks[bssid]
        return

    def on_popup_shown(self, combo_box, params):
        if combo_box.get_property("popup-shown"):
            logging.debug("Dropdown popup show")
            self.last_drop_time = datetime.now()
        else:
            elapsed = (datetime.now() - self.last_drop_time).total_seconds()
            if elapsed < 0.2:
                logging.debug(f"Dropdown closed too fast ({elapsed}s)")
                GLib.timeout_add(50, combo_box.popup)
                return
            logging.debug("Dropdown popup close")

    def show_add_network(self, widget, ssid):
        if self.show_add:
            return

        for child in self.content.get_children():
            self.content.remove(child)

        if "add_network" in self.labels:
            del self.labels['add_network']

        eap_method = Gtk.ComboBoxText(hexpand=True)
        eap_method.connect("notify::popup-shown", self.on_popup_shown)
        for method in ("peap", "ttls", "pwd", "leap", "md5"):
            eap_method.append(method, method.upper())
        self.labels['network_eap_method'] = eap_method
        eap_method.set_active(0)

        phase2 = Gtk.ComboBoxText(hexpand=True)
        phase2.connect("notify::popup-shown", self.on_popup_shown)
        for method in ("mschapv2", "gtc", "pap", "chap", "mschap", "disabled"):
            phase2.append(method, method.upper())
        self.labels['network_phase2'] = phase2
        phase2.set_active(0)

        auth_selection_box = Gtk.Box(no_show_all=True)
        auth_selection_box.add(self.labels['network_eap_method'])
        auth_selection_box.add(self.labels['network_phase2'])

        self.labels['network_identity'] = Gtk.Entry(hexpand=True, no_show_all=True)
        self.labels['network_identity'].connect("touch-event", self._screen.show_keyboard)
        self.labels['network_identity'].connect("button-press-event", self._screen.show_keyboard)

        self.labels['network_psk'] = Gtk.Entry(hexpand=True)
        self.labels['network_psk'].connect("activate", self.add_new_network, ssid)
        self.labels['network_psk'].connect("touch-event", self._screen.show_keyboard)
        self.labels['network_psk'].connect("button-press-event", self._screen.show_keyboard)

        save = self._gtk.Button("sd", _("Save"), "color3")
        save.set_hexpand(False)
        save.connect("clicked", self.add_new_network, ssid)

        user_label = Gtk.Label(label=_("User"), hexpand=False, no_show_all=True)
        auth_grid = Gtk.Grid()
        auth_grid.attach(user_label, 0, 0, 1, 1)
        auth_grid.attach(self.labels['network_identity'], 1, 0, 1, 1)
        auth_grid.attach(Gtk.Label(label=_("Password"), hexpand=False), 0, 1, 1, 1)
        auth_grid.attach(self.labels['network_psk'], 1, 1, 1, 1)
        auth_grid.attach(save, 2, 0, 1, 2)

        if "802.1x" in self.sdbus_nm.get_security_type(ssid):
            user_label.show()
            self.labels['network_eap_method'].show()
            self.labels['network_phase2'].show()
            self.labels['network_identity'].show()
            auth_selection_box.show()

        self.labels['add_network'] = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL, spacing=5, valign=Gtk.Align.CENTER,
            hexpand=True, vexpand=True
        )
        self.labels['add_network'].add(Gtk.Label(label=_("Connecting to %s") % ssid))
        self.labels['add_network'].add(auth_selection_box)
        self.labels['add_network'].add(auth_grid)
        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.labels['add_network'])
        self.content.add(scroll)
        self.labels['network_psk'].grab_focus_without_selecting()
        self.content.show_all()
        self.show_add = True

    def update_all_networks(self):
        self.interface = self.sdbus_nm.get_primary_interface()
        self.labels['ip'].set_text(f"IP: {self.sdbus_nm.get_ip_address()}")
        nets = self.sdbus_nm.get_networks()

        current_bssids = {net['BSSID'] for net in nets if 'BSSID' in net}
        to_remove = [bssid for bssid in self.network_rows if bssid not in current_bssids]
        for bssid in to_remove:
            self.remove_network_from_list(bssid)

        for net in nets:
            bssid = net.get('BSSID')
            if not bssid:
                continue
            if bssid not in self.network_rows:
                self.add_network(bssid)
            self.update_network_info(net)

        for i, net in enumerate(nets):
            bssid = net.get('BSSID')
            widget = self.network_rows.get(bssid)
            if widget:
                self.network_list.reorder_child(widget, i)
        self.network_list.show_all()
        return True

    def update_network_info(self, net):
        if net['BSSID'] not in self.network_rows.keys() or net['BSSID'] not in self.networks:
            logging.info(f"Unknown SSID {net['SSID']}")
            return
        chan = _("Channel") + f' {net["channel"]}'
        self.networks[net['BSSID']]['icon'].set_from_pixbuf(self.get_signal_strength_icon(net["signal_level"]))
        self.networks[net['BSSID']]['info'].set_markup(
            "<small>"
            f"{net['frequency']} Ghz  {chan}  {net['signal_level']} %\n"
            f"{net['security']}\n"
            f"{net['BSSID']}"
            "</small>"
        )

    def get_signal_strength_icon(self, signal_level):
        # networkmanager uses percentage not dbm
        if signal_level > 75:
            return self.wifi_signal_icons['excellent']
        elif signal_level > 60:
            return self.wifi_signal_icons['good']
        elif signal_level > 30:
            return self.wifi_signal_icons['fair']
        else:
            return self.wifi_signal_icons['weak']

    def update_single_network_info(self):
        self.labels['networkinfo'].set_markup(
            f'<b>{self.interface}</b>\n\n'
            + '<b>' + _("Hostname") + f':</b> {os.uname().nodename}\n'
            f'<b>IPv4:</b> {self.sdbus_nm.get_ip_address()}\n'
        )
        self.labels['networkinfo'].show_all()
        return True

    def reload_networks(self, widget=None):
        self.deactivate()
        del self.network_rows
        self.network_rows = {}
        for child in self.network_list.get_children():
            self.network_list.remove(child)
        if self.sdbus_nm is not None and self.sdbus_nm.wifi:
            if widget:
                self._gtk.Button_busy(widget, True)
            self.sdbus_nm.rescan()
            self.load_networks()
        self.activate()

    def activate(self):
        if self.sdbus_nm is None:
            return
        if self.update_timeout is None:
            if self.sdbus_nm.wifi:
                if self.reload_button.get_sensitive():
                    self._gtk.Button_busy(self.reload_button, True)
                    self.sdbus_nm.rescan()
                    self.load_networks()
                self.update_all_networks()
                self.update_timeout = GLib.timeout_add_seconds(5, self.update_all_networks)
            else:
                self.update_single_network_info()
                self.update_timeout = GLib.timeout_add_seconds(5, self.update_single_network_info)

    def deactivate(self):
        if self.sdbus_nm is None:
            return
        if self.update_timeout is not None:
            GLib.source_remove(self.update_timeout)
            self.update_timeout = None
        self.sdbus_nm.set_connection_monitoring(False)

    def toggle_wifi(self, switch, gparams):
        enable = switch.get_active()
        logging.info(f"WiFi {enable}")
        self.sdbus_nm.toggle_wifi(enable)
        if enable:
            self.reload_button.show()
            self.reload_networks()
        else:
            self.reload_button.hide()



import configparser
import copy
import gettext
import json
import locale
import logging
import os
import pathlib
import re
from io import StringIO

SCREEN_BLANKING_OPTIONS = [
    60,     # 1 Minute
    120,    # 2 Minutes
    300,    # 5 Minutes
    900,    # 15 Minutes
    1800,   # 30 Minutes
    3600,   # 1 Hour
    7200,   # 2 Hours
    14400,  # 4 Hours
]

klipperscreendir = pathlib.Path(__file__).parent.resolve().parent
home = os.path.expanduser("~/")
printer_data_config = os.path.join(home, "printer_data", "config")
xdg_config = os.path.join(home, ".config", "KlipperScreen")


class ConfigError(Exception):
    pass


class KlipperScreenConfig:
    config = None
    configfile_name = "KlipperScreen.conf"
    do_not_edit_line = "#~# --- Do not edit below this line. This section is auto generated --- #~#"
    do_not_edit_prefix = "#~#"

    def __init__(self, configfile, screen=None):
        self.lang_list = None
        self.errors = []
        self.default_config_path = os.path.join(klipperscreendir, "config", "defaults.conf")
        self.config = configparser.ConfigParser()
        self.config_path = self.get_config_file_location(configfile)
        logging.debug(f"Config path location: {self.config_path}")
        self.user_cfg = None
        self.lang = None
        self.langs = {}

        try:
            # parse defaults.conf
            self.config.read(self.default_config_path)
            includes = [i[8:] for i in self.config.sections() if i.startswith("include ")]
            for include in includes:
                directory = "/".join(self.default_config_path.split("/")[:-1])
                self._include_config(self.config, directory, include, log=False)
            self.validate_config(self.config)  # In case a user altered defaults
            # parse KlipperScreen.conf
            if self.config_path != self.default_config_path:
                user_def, saved_def = self.separate_saved_config(self.config_path)
                self.user_cfg = configparser.ConfigParser()
                self.user_cfg.read_string(user_def)

                includes = [i[8:] for i in self.user_cfg.sections() if i.startswith("include ")]
                for include in includes:
                    directory = "/".join(self.config_path.split("/")[:-1])
                    self._include_config(self.user_cfg, directory, include)

                if not self.user_cfg.getboolean('main', "use_default_menu", fallback=True):
                    self.exclude_menu_from_config(self.config)

                if any(sec.startswith("preheat") for sec in self.user_cfg.sections()):
                    self.exclude_preheat_from_config(self.config)

                self.log_config(self.user_cfg)
                if self.validate_config(self.user_cfg, string=user_def):
                    merged_user_text = self._build_config_string(self.user_cfg)
                    self.config.read_string(merged_user_text)  # Merge user config
                # Parse configs saved from the UI
                if saved_def is not None:
                    auto_gen = configparser.ConfigParser()
                    auto_gen.read_string(saved_def)
                    if self.validate_config(auto_gen, string=saved_def, remove=True):
                        self.config.read_string(saved_def)  # Merge ui saved config
                        logging.info(f"\n====== Saved ======\n{saved_def}\n")
            # This is the final config
            # self.log_config(self.config)
        except KeyError as Kerror:
            msg = f"Error reading config: {self.config_path}\n{Kerror}"
            logging.exception(msg)
            self.errors.append(msg)
            raise ConfigError(msg) from Kerror
        except ValueError as Verror:
            msg = f"Invalid Value in the config:\n{Verror}"
            logging.exception(msg)
            self.errors.append(msg)
        except Exception as e:
            msg = f"Unknown error with the config:\n{e}"
            logging.exception(msg)
            self.errors.append(msg)

        printers = [i for i in self.config.sections() if i.startswith("printer ")]
        if not printers:
            printers.append("Printer Printer")
        self.printers = [
            {printer[8:]: {
                "moonraker_host": self.config.get(printer, "moonraker_host", fallback="127.0.0.1"),
                "moonraker_port": self.config.get(printer, "moonraker_port", fallback="7125"),
                "moonraker_path": self.config.get(printer, "moonraker_path", fallback='').strip('/'),
                "moonraker_ssl": self.config.getboolean(printer, "moonraker_ssl", fallback=None),
                "moonraker_api_key": self.config.get(printer, "moonraker_api_key", fallback="").replace('"', '')
            }} for printer in printers
        ]

        conf_printers_debug = copy.deepcopy(self.printers)
        for printer in conf_printers_debug:
            name = list(printer)[0]
            item = conf_printers_debug[conf_printers_debug.index(printer)]
            if item[name]['moonraker_api_key'] != "":
                item[name]['moonraker_api_key'] = "redacted"
        logging.debug(f"Configured printers: {json.dumps(conf_printers_debug, indent=2)}")

        self.create_translations()
        self._create_configurable_options(screen)

    def create_translations(self):
        lang_path = os.path.join(klipperscreendir, "ks_includes", "locales")
        self.lang_list = [d for d in os.listdir(lang_path) if not os.path.isfile(os.path.join(lang_path, d))]
        self.lang_list.sort()
        for lng in self.lang_list:
            self.langs[lng] = gettext.translation('KlipperScreen', localedir=lang_path, languages=[lng], fallback=True)

        lang = self.get_main_config().get("language", "system_lang")
        logging.debug(f"Selected lang: {lang} OS lang: {locale.getlocale()[0]}")
        self.install_language(lang)

    def find_language(self, lang):
        if lang in (None, "system_lang"):
            sys_lang = locale.getlocale()[0]
            if sys_lang is None or len(sys_lang) < 2:
                return "en"
            if sys_lang in self.lang_list:
                return sys_lang
            for language in self.lang_list:
                if sys_lang.startswith(language):
                    return language
        return next((language for language in self.lang_list if lang.startswith(language)), "en")

    def install_language(self, lang):
        if lang not in self.lang_list:
            lang = self.find_language(lang)
        logging.info(f"Using lang {lang}")
        self.lang = self.langs[lang]
        self.lang.install(names=['gettext', 'ngettext'])

    def validate_config(self, config, string="", remove=False):
        valid = True
        if string:
            msg = "Section headers have extra information after brackets possible newline issue:"
            for line in string.split('\n'):
                if re.match(r".+\].", line):
                    logging.error(line)
                    self.errors.append(f'{msg}\n\n{line}')
                    return False
        for section in config:
            if section == 'DEFAULT' or section.startswith('include '):
                # Do not validate 'DEFAULT' or 'include*' sections
                continue
            bools = strs = numbers = ()
            if section == 'main':
                bools = (
                    'invert_x', 'invert_y', 'invert_z', '24htime', 'only_heaters', 'show_cursor', 'confirm_estop',
                    'autoclose_popups', 'use_dpms', 'use_default_menu', 'side_macro_shortcut', 'use-matchbox-keyboard',
                    'show_heater_power', "show_scroll_steppers", "auto_open_extrude", "start_locked",
                    "keyboard_navigation"
                )
                strs = (
                    'default_printer', 'language', 'print_sort_dir', 'theme', 'screen_blanking_printing', 'font_size',
                    'print_estimate_method', 'screen_blanking', "screen_on_devices", "screen_off_devices", 'print_view',
                    "lock_password"
                )
                numbers = (
                    'job_complete_timeout', 'job_error_timeout', 'move_speed_xy', 'move_speed_z',
                    'print_estimate_compensation', 'width', 'height', 'autolock_timeout',

                )
            elif section.startswith('printer '):
                bools = (
                    'invert_x', 'invert_y', 'invert_z', 'moonraker_ssl',
                )
                strs = (
                    'moonraker_api_key', 'moonraker_host', 'moonraker_path', 'titlebar_name_type',
                    'screw_positions', 'power_devices', 'titlebar_items', 'hidden_sensors', 'z_babystep_values',
                    'extrude_distances', 'extrude_speeds', 'move_distances', 'zcalibrate_custom_commands'
                )
                numbers = (
                    'moonraker_port', 'move_speed_xy', 'move_speed_z', 'screw_rotation',
                    'calibrate_x_position', 'calibrate_y_position', 'spool_low_limit',
                )
            elif section.startswith('preheat '):
                strs = ('gcode', '')
                numbers = [f'{option}' for option in config[section] if option != 'gcode']
            elif section.startswith('menu '):
                strs = ('name', 'icon', 'panel', 'method', 'params', 'enable', 'confirm', 'style')
            elif section.startswith('graph')\
                    or section.startswith('displayed_macros')\
                    or section.startswith('spoolman'):
                bools = [f'{option}' for option in config[section]]
            else:
                self.errors.append(f'Section [{section}] not recognized')

            for key in config[section]:
                if key not in bools and key not in strs and key not in numbers:
                    msg = f'Option "{key}" not recognized for section "[{section}]"'
                    if section.startswith("menu"):
                        msg = (
                            msg + "\n\n"
                            + "KlipperScreen menu uses different options than Klipper\n"
                            + "Syntax meant for text/graphical LCDs (usually blue)\n"
                            + "will not work with KlipperScreen\n"
                            + "please consult KlipperScreen documentation."
                        )
                    if key == "camera_url":
                        msg = (
                            "camera_url has been deprecated in favor of moonraker cameras\n\n"
                            + "https://moonraker.readthedocs.io/en/latest/configuration/#webcam\n\n"
                            + "remove camera_url from KlipperScreen config file"
                        )
                    if remove:
                        # This should only be called for the auto-generated section
                        self.config.remove_option(section, key)
                    else:
                        self.errors.append(msg)
                elif key in numbers and not self.is_float(config[section][key]) \
                        or key in bools and not self.is_bool(config[section][key]):
                    msg = (
                        f'Unable to parse "{key}" from [{section}]\n'
                        f'Expected a {"number" if key in numbers else "boolean"} but got: {config[section][key]}'
                    )
                    self.errors.append(msg)
                    logging.error('Invalid configuration detected !!!')
                    valid = False
        return valid

    @staticmethod
    def is_float(element):
        try:
            float(element)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_bool(element):
        return element in ["False", "false", "True", "true"]

    def get_errors(self):
        return "".join(f'{error}\n\n' for error in self.errors)

    def _create_configurable_options(self, screen):

        self.configurable_options = [
            {"theme": {
                "section": "main", "name": _("Icon Theme"), "type": "dropdown",
                "tooltip": _("Changes how the interface looks"),
                "value": "z-bolt", "callback": screen.change_theme, "options": [
                    {"name": "Z-bolt" + " " + _("(default)"), "value": "z-bolt"}]}},
            {"print_estimate_method": {
                "section": "main", "name": _("Estimated Time Method"), "type": "dropdown",
                "tooltip": _("Changes how the time remaining is calculated"),
                "value": "auto", "options": [
                    {"name": _("Auto") + " " + _("(default)"), "value": "auto"},
                    {"name": _("File"), "value": "file"},
                    {"name": _("Filament Used"), "value": "filament"},
                    {"name": _("Slicer"), "value": "slicer"}]}},
            {"screen_blanking": {
                "section": "main", "name": _("Screen Power Off Time"), "type": "dropdown",
                "tooltip": _("Timeout for screen black-out or power-off"),
                "value": "3600", "callback": screen.set_screenblanking_timeout, "options": [
                    {"name": _("Never"), "value": "off"}]
            }},
            {"screen_blanking_printing": {
                "section": "main", "name": _("Screen Power Off Time") + " (" + _("Printing") + ")",
                "type": "dropdown",
                "tooltip": _("Timeout for screen black-out or power-off during printing"),
                "value": "3600", "callback": screen.set_screenblanking_printing_timeout, "options": [
                    {"name": _("Never"), "value": "off"}]
            }},
            {"24htime": {"section": "main", "name": _("24 Hour Time"), "type": "binary",
                         "tooltip": _("Disable for 12hs with am / pm"),
                         "value": "True"}},
            {"side_macro_shortcut": {
                "section": "main", "name": _("Macro shortcut on sidebar"), "type": "binary",
                "value": "False", "callback": screen.toggle_shortcut}},
            {"font_size": {
                "section": "main", "name": _("Font Size"), "type": "dropdown",
                "tooltip": _("Inversely affects the icon size"),
                "value": "medium", "callback": screen.restart_ks, "options": [
                    {"name": _("Small"), "value": "small"},
                    {"name": _("Medium") + " " + _("(default)"), "value": "medium"},
                    {"name": _("Large"), "value": "large"},
                    {"name": _("Extra Large"), "value": "extralarge"},
                    {"name": _("Maximum"), "value": "max"}]}},
            {"confirm_estop": {"section": "main", "name": _("Confirm Emergency Stop"), "type": "binary",
                               "tooltip": _("Asks for confirmation before stopping"),
                               "value": "False"}},
            {"only_heaters": {"section": "main", "name": _("Hide sensors in Temp."), "type": "binary",
                              "tooltip": _("Show only devices that are able to be set"),
                              "value": "False", "callback": screen.reload_panels}},
            {"use_dpms": {"section": "main", "name": _("Screen DPMS"), "type": "binary",
                          "tooltip": _("Enable screen power management") + "\n"
                                     + _("Not all screens support this"),
                          "value": "True", "callback": screen.set_dpms}},
            {"autoclose_popups": {"section": "main", "name": _("Auto-close notifications"), "type": "binary",
                                  "tooltip": _("Close messages after a timeout"),
                                  "value": "True"}},
            {"show_heater_power": {"section": "main", "name": _("Show Heater Power"), "type": "binary",
                                   "tooltip": _("Current percentage and graph line"),
                                   "value": "False", "callback": screen.reload_panels}},
            {"show_scroll_steppers": {"section": "main", "name": _("Show Scrollbars Buttons"), "type": "binary",
                                      "tooltip": _("Useful for un-responsive touchscreens"),
                                      "value": "False", "callback": screen.reload_panels}},
            {"auto_open_extrude": {"section": "main", "name": _("Auto-open Extrude On Pause"), "type": "binary",
                                   "value": "True", "callback": screen.reload_panels}},
            {"show_cursor": {"section": "main", "name": _("Show cursor"), "type": "binary",
                             "tooltip": _("For mouse control or to verify touchscreen accuracy"),
                             "value": "False", "callback": screen.update_cursor}},
            # {"": {"section": "main", "name": _(""), "type": ""}}
        ]

        # Options that are in panels and shouldn't be added to the main settings
        panel_options = [
            {"invert_x": {"section": "main", "name": _("Invert X"), "type": None, "value": "False"}},
            {"invert_y": {"section": "main", "name": _("Invert Y"), "type": None, "value": "False"}},
            {"invert_z": {"section": "main", "name": _("Invert Z"), "type": None, "value": "False"}},
            {"move_speed_xy": {"section": "main", "name": _("XY Move Speed (mm/s)"), "type": None, "value": "50"}},
            {"move_speed_z": {"section": "main", "name": _("Z Move Speed (mm/s)"), "type": None, "value": "10"}},
            {"print_sort_dir": {"section": "main", "type": None, "value": "name_asc"}},
            {"print_view": {"section": "main", "type": None, "value": "thumbs"}},
            {"language": {"section": "main", "name": _("Language"), "type": None, "value": "system_lang"}},
        ]

        self.configurable_options.extend(panel_options)

        i0 = i1 = i2 = None
        for i, option in enumerate(self.configurable_options):
            if list(option)[0] == "theme":
                i0 = i
            elif list(option)[0] == "screen_blanking":
                i1 = i
            elif list(option)[0] == "screen_blanking_printing":
                i2 = i
            if i0 and i1 and i2:
                break

        t_path = os.path.join(klipperscreendir, 'styles')
        themes = [
            d for d in os.listdir(t_path)
            if (not os.path.isfile(os.path.join(t_path, d))
                and d not in ("z-bolt", "printers"))
        ]
        themes.sort()

        for theme in themes:
            self.configurable_options[i0]['theme']['options'].append({"name": theme, "value": theme})

        for num in SCREEN_BLANKING_OPTIONS:
            hour = num // 3600
            minute = num // 60
            if hour > 0:
                name = f'{hour} ' + ngettext("hour", "hours", hour)
            else:
                name = f'{minute} ' + ngettext("minute", "minutes", minute)
            self.configurable_options[i1]['screen_blanking']['options'].append({
                "name": name,
                "value": f"{num}"
            })
            self.configurable_options[i2]['screen_blanking_printing']['options'].append({
                "name": name,
                "value": f"{num}"
            })

        for item in self.configurable_options:
            name = list(item)[0]
            vals = item[name]
            if vals['section'] not in self.config.sections():
                self.config.add_section(vals['section'])
            if name not in list(self.config[vals['section']]):
                self.config.set(vals['section'], name, vals['value'])

    def exclude_menu_from_config(self, config):
        logging.info("Using custom menu, removing default menu entries.")
        exclude_list = ('menu __main', 'menu __print', 'menu __splashscreen')
        self.exclude_from_config(config, exclude_list)

    def exclude_preheat_from_config(self, config):
        exclude_list = ('preheat',)
        self.exclude_from_config(config, exclude_list)

    def exclude_from_config(self, config, exclude_list):
        for prefix in exclude_list:
            if any(sec.startswith(prefix) for sec in config.sections()):
                for key in list(self.config.keys()):
                    if key.startswith(prefix):
                        del self.config[key]

    def _include_config(self,
                        config_to_integrate: configparser.ConfigParser,
                        base_dir: str,
                        filepath: str,
                        log: bool = True) -> None:
        # ---------- Resolve absolute path ----------
        if os.path.isabs(filepath):
            full_path = filepath
        else:
            full_path = os.path.join(base_dir, filepath)

        files_to_process = []

        if '*' in full_path:
            parent_dir = os.path.dirname(full_path)
            pattern = re.escape(os.path.basename(full_path)).replace(r'\*', r'.*')
            regex = re.compile(f'^{pattern}$')

            try:
                for fname in os.listdir(parent_dir):
                    if regex.match(fname):
                        files_to_process.append(os.path.join(parent_dir, fname))
            except Exception as e:
                logging.error(f"Config Error: unable to list directory {parent_dir}: {e}")
                return
        else:
            if not os.path.isfile(full_path):
                logging.error(f"Config Error: include file '{full_path}' does not exist")
                return
            files_to_process.append(full_path)

        for fp in files_to_process:
            try:
                with open(fp, 'r', encoding='utf-8') as f:
                    cfg_text = f.read()
            except Exception as e:
                logging.error(f"Config Error: failed to read '{fp}': {e}")
                continue

            temp_cfg = configparser.ConfigParser()
            try:
                temp_cfg.read_string(cfg_text)
            except Exception as e:
                logging.error(f"Config Error: failed to parse '{fp}': {e}")
                continue

            # nested includes
            nested_includes = [s[8:] for s in temp_cfg.sections() if s.startswith("include ")]
            for inc_file in nested_includes:
                self._include_config(
                    config_to_integrate=config_to_integrate,
                    base_dir=os.path.dirname(fp),
                    filepath=inc_file,
                    log=False
                )
            if self.validate_config(config_to_integrate, string=cfg_text):
                config_to_integrate.read_string(cfg_text)
                if log:
                    self.log_config(config_to_integrate)

    def separate_saved_config(self, config_path):
        user_def = []
        saved_def = []
        found_saved = False
        if not os.path.exists(config_path):
            return ["", None]
        with open(config_path) as file:
            for line in file:
                line = line.replace('\n', '')
                if line == self.do_not_edit_line:
                    found_saved = True
                    saved_def = []
                    continue
                if found_saved is False:
                    user_def.append(line.replace('\n', ''))
                elif line.startswith(self.do_not_edit_prefix):
                    saved_def.append(line[(len(self.do_not_edit_prefix) + 1):])
        return ["\n".join(user_def), None if saved_def is None else "\n".join(saved_def)]

    @staticmethod
    def check_path_exists(base_dir, filename):
        for name in (filename, filename.lower()):
            full_path = os.path.join(base_dir, name)
            if os.path.exists(full_path):
                return full_path
        return None

    def get_config_file_location(self, file):
        # Passed config (-c) by default is blank
        logging.info(f"Passed config (-c): {file}")
        if file not in (".", "..") and os.path.exists(file):
            return file

        # List of directories to search for the config file
        directories = [printer_data_config, xdg_config, klipperscreendir]

        for directory in directories:
            path = self.check_path_exists(directory, self.configfile_name)
            if path:
                return path

        # fallback
        return self.default_config_path

    def get_config(self):
        return self.config

    def get_configurable_options(self):
        return self.configurable_options

    def get_lang(self):
        return self.lang

    def get_main_config(self):
        return self.config['main']

    def get_menu_items(self, menu="__main", subsection=""):
        if subsection != "":
            subsection = f"{subsection} "
        index = f"menu {menu} {subsection}"
        items = [i[len(index):] for i in self.config.sections() if i.startswith(index)]
        menu_items = []
        for item in items:
            split = item.split()
            if len(split) == 1:
                menu_items.append(self._build_menu_item(menu, index + item))

        return menu_items

    def get_menu_name(self, menu="__main", subsection=""):
        name = f"menu {menu} {subsection}" if subsection != "" else f"menu {menu}"
        return False if name not in self.config else self.config[name].get('name')

    def get_preheat_options(self):
        index = "preheat "
        items = [i[len(index):] for i in self.config.sections() if i.startswith(index)]
        return {item: self._build_preheat_item(index + item) for item in items}

    def _build_preheat_item(self, name):
        if name not in self.config:
            return False
        cfg = self.config[name]
        return {opt: cfg.get("gcode", None) if opt == "gcode" else cfg.getfloat(opt, None) for opt in cfg}

    def get_printer_config(self, name):
        if not name.startswith("printer "):
            name = f"printer {name}"

        return None if name not in self.config else self.config[name]

    def get_printers(self):
        return self.printers

    def save_user_config_options(self):
        save_config = configparser.ConfigParser()
        for item in self.configurable_options:
            name = list(item)[0]
            opt = item[name]
            curval = self.config[opt['section']].get(name)
            if curval != opt["value"] or (
                    self.user_cfg is not None and opt['section'] in self.user_cfg.sections() and
                    self.user_cfg[opt['section']].get(name, None) not in (None, curval)):
                if opt['section'] not in save_config.sections():
                    save_config.add_section(opt['section'])
                save_config.set(opt['section'], name, str(curval))

        extra_sections = [i for i in self.config.sections() if i.startswith("displayed_macros")]
        extra_sections.extend([i for i in self.config.sections() if i.startswith("graph")])
        extra_sections.extend([i for i in self.config.sections() if i.startswith("spoolman")])
        for section in extra_sections:
            for item in self.config.options(section):
                value = self.config[section].getboolean(item, fallback=True)
                if value is False or (self.user_cfg is not None and
                                      section in self.user_cfg.sections() and
                                      self.user_cfg[section].getboolean(item, fallback=True) is False and
                                      self.user_cfg[section].getboolean(item, fallback=True) != value):
                    if section not in save_config.sections():
                        save_config.add_section(section)
                    save_config.set(section, item, str(value))

        save_output = self._build_config_string(save_config).split("\n")
        for i in range(len(save_output)):
            save_output[i] = f"{self.do_not_edit_prefix} {save_output[i]}"

        if self.config_path == self.default_config_path:
            user_def = ""
            saved_def = None
        else:
            user_def, saved_def = self.separate_saved_config(self.config_path)

        contents = (f"{user_def}\n"
                    f"{self.do_not_edit_line}\n"
                    f"{self.do_not_edit_prefix}\n"
                    + '\n'.join(save_output) + f"\n"
                                               f"{self.do_not_edit_prefix}\n")

        if self.config_path != self.default_config_path:
            filepath = self.config_path
        else:
            if os.path.exists(printer_data_config):
                filepath = os.path.join(printer_data_config, self.configfile_name)
            else:
                try:
                    if not os.path.exists(xdg_config):
                        pathlib.Path(xdg_config).mkdir(parents=True, exist_ok=True)
                    filepath = os.path.join(xdg_config, self.configfile_name)
                except Exception as e:
                    logging.error(e)
                    filepath = klipperscreendir
            logging.info(f'Creating a new config file in {filepath}')
        try:
            with open(filepath, 'w') as file:
                file.write(contents)
        except Exception as e:
            logging.error(f"Error writing configuration file in {filepath}:\n{e}")

    def set(self, section, name, value):
        self.config.set(section, name, str(value))

    def log_config(self, config):
        sensitive_keys = [
            r'(moonraker_api_key\s*(?:=|:)\s*)\S+',
            r'(lock_password\s*(?:=|:)\s*)\S+',
        ]
        config_str = self._build_config_string(config)
        for pattern in sensitive_keys:
            config_str = re.sub(pattern, r'\1[redacted]', config_str)
        lines = [
            "===== Config File =====",
            config_str,
            "======================="
        ]
        logging.info("\n".join(lines))

    @staticmethod
    def _build_config_string(config):
        sfile = StringIO()
        config.write(sfile)
        sfile.seek(0)
        return sfile.read().strip()

    def _build_menu_item(self, menu, name):
        if name not in self.config:
            return False
        cfg = self.config[name]
        item = {
            "name": cfg.get("name"),
            "icon": cfg.get("icon", None),
            "panel": cfg.get("panel", None),
            "method": cfg.get("method", None),
            "confirm": cfg.get("confirm", None),
            "enable": cfg.get("enable", "True"),
            "params": cfg.get("params", "{}"),
            "style": cfg.get("style", None)
        }

        return {name[(len(menu) + 6):]: item}



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):

    def __init__(self, screen, title):
        title = title or _("Retraction")
        super().__init__(screen, title)
        self.options = None
        self.grid = Gtk.Grid()
        self.values = {}
        self.list = {}
        conf = self._printer.get_config_section("firmware_retraction")

        retract_length = float(conf['retract_length']) if 'retract_length' in conf else 0
        retract_speed = int(float((conf['retract_speed']))) if 'retract_speed' in conf else 20
        unretract_extra_length = float(conf['unretract_extra_length']) if 'unretract_extra_length' in conf else 0
        unretract_speed = int(float((conf['unretract_speed']))) if 'unretract_speed' in conf else 10
        maxlength = retract_length * 1.2 if retract_length >= 6 else 6
        maxspeed = retract_speed * 1.2 if retract_speed >= 100 else 100

        self.options = [
            {"name": _("Retraction Length"),
             "units": _("mm"),
             "option": "retract_length",
             "value": retract_length,
             "digits": 2,
             "maxval": maxlength},
            {"name": _("Retraction Speed"),
             "units": _("mm/s"),
             "option": "retract_speed",
             "value": retract_speed,
             "digits": 0,
             "maxval": maxspeed},
            {"name": _("Unretract Extra Length"),
             "units": _("mm"),
             "option": "unretract_extra_length",
             "value": unretract_extra_length,
             "digits": 2,
             "maxval": maxlength},
            {"name": _("Unretract Speed"),
             "units": _("mm/s"),
             "option": "unretract_speed",
             "value": unretract_speed,
             "digits": 0,
             "maxval": maxspeed}
        ]

        for opt in self.options:
            self.add_option(opt['option'], opt['name'], opt['units'], opt['value'], opt['digits'], opt["maxval"])

        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.grid)

        self.content.add(scroll)
        self.content.show_all()

    def activate(self):
        self._screen._ws.klippy.gcode_script("GET_RETRACTION")

    def process_update(self, action, data):
        if action == "notify_status_update" and "firmware_retraction" in data:
            for opt in self.list:
                if opt in data["firmware_retraction"]:
                    self.update_option(opt, data["firmware_retraction"][opt])

    def update_option(self, option, value):
        if option not in self.list:
            return

        if self.list[option]['scale'].has_grab():
            return

        self.values[option] = float(value)
        # Infinite scale
        for opt in self.options:
            if opt['option'] == option:
                if self.values[option] > opt["maxval"] * .75:
                    self.list[option]['adjustment'].set_upper(self.values[option] * 1.5)
                else:
                    self.list[option]['adjustment'].set_upper(opt["maxval"])
                break
        self.list[option]['scale'].set_value(self.values[option])
        self.list[option]['scale'].disconnect_by_func(self.set_opt_value)
        self.list[option]['scale'].connect("button-release-event", self.set_opt_value, option)

    def add_option(self, option, optname, units, value, digits, maxval):
        logging.info(f"Adding option: {option}")

        name = Gtk.Label(
            hexpand=True, vexpand=True, halign=Gtk.Align.START, valign=Gtk.Align.CENTER,
            wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        name.set_markup(f"<big><b>{optname}</b></big> ({units})")
        minimum = 1 if option in ["retract_speed", "unretract_speed"] else 0
        self.values[option] = value
        # adj (value, lower, upper, step_increment, page_increment, page_size)
        adj = Gtk.Adjustment(value, minimum, maxval, 1, 5, 0)
        scale = Gtk.Scale(adjustment=adj, digits=digits, hexpand=True,
                          has_origin=True)
        scale.get_style_context().add_class("option_slider")
        scale.connect("button-release-event", self.set_opt_value, option)

        reset = self._gtk.Button("refresh", style="color1")
        reset.connect("clicked", self.reset_value, option)
        reset.set_hexpand(False)

        item = Gtk.Grid()
        item.attach(name, 0, 0, 2, 1)
        item.attach(scale, 0, 1, 1, 1)
        item.attach(reset, 1, 1, 1, 1)

        self.list[option] = {
            "row": item,
            "scale": scale,
            "adjustment": adj,
        }

        pos = sorted(self.list).index(option)
        self.grid.attach(self.list[option]['row'], 0, pos, 1, 1)
        self.grid.show_all()

    def reset_value(self, widget, option):
        for x in self.options:
            if x["option"] == option:
                self.update_option(option, x["value"])
        self.set_opt_value(None, None, option)

    def set_opt_value(self, widget, event, opt):
        value = self.list[opt]['scale'].get_value()

        if opt == "retract_speed":
            self._screen._ws.klippy.gcode_script(f"SET_RETRACTION RETRACT_SPEED={value}")
        elif opt == "retract_length":
            self._screen._ws.klippy.gcode_script(f"SET_RETRACTION RETRACT_LENGTH={value}")
        elif opt == "unretract_extra_length":
            self._screen._ws.klippy.gcode_script(f"SET_RETRACTION UNRETRACT_EXTRA_LENGTH={value}")
        elif opt == "unretract_speed":
            self._screen._ws.klippy.gcode_script(f"SET_RETRACTION UNRETRACT_SPEED={value}")



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, Gtk


class ObjectMap(Gtk.DrawingArea):
    def __init__(self, screen, printer, font_size):
        super().__init__()
        self._screen = screen
        self.set_hexpand(True)
        self.set_vexpand(True)
        # self.get_style_context().add_class('objectmap')
        self.printer = printer
        self.max_length = 0
        self.connect('draw', self.draw_graph)
        self.add_events(Gdk.EventMask.TOUCH_MASK)
        self.add_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        self.connect('button_press_event', self.event_cb)
        self.font_size = round(font_size * 0.75)
        self.font_spacing = round(self.font_size * 1.5)
        self.margin_left = round(self.font_size * 2.75)
        self.margin_right = 15
        self.margin_top = 10
        self.margin_bottom = self.font_size * 2
        self.objects = self.printer.get_stat("exclude_object", "objects")
        self.current_object = self.printer.get_stat("current_object", "current_object")
        self.excluded_objects = self.printer.get_stat("exclude_object", "excluded_objects")
        self.min_x = self.min_y = 99999999
        self.max_x = self.max_y = 0

    def x_graph_to_bed(self, width, gx):
        return (((gx - self.margin_left) * (self.max_x - self.min_x))
                / (width - self.margin_left - self.margin_right)) + self.min_x

    def y_graph_to_bed(self, height, gy):
        return ((1 - ((gy - self.margin_top) / (height - self.margin_top - self.margin_bottom)))
                * (self.max_y - self.min_y)) + self.min_y

    def event_cb(self, da, ev):
        # Convert coordinates from screen-graph to bed
        x = self.x_graph_to_bed(da.get_allocated_width(), ev.x)
        y = self.y_graph_to_bed(da.get_allocated_height(), ev.y)
        logging.info(f"Touched GRAPH {ev.x:.0f},{ev.y:.0f} BED: {x:.0f},{y:.0f}")

        for obj in self.objects:
            obj_min_x = obj_max_x = obj["polygon"][0][0]
            obj_min_y = obj_max_y = obj["polygon"][0][1]
            for point in obj["polygon"]:
                obj_min_x = min(obj_min_x, point[0])
                obj_min_y = min(obj_min_y, point[1])
                obj_max_x = max(obj_max_x, point[0])
                obj_max_y = max(obj_max_y, point[1])
            if obj_min_x < x < obj_max_x and obj_min_y < y < obj_max_y:
                logging.info(f"TOUCHED object it's: {obj['name']}")
                if obj['name'] not in self.excluded_objects:
                    self.exclude_object(obj['name'])
                break

    def exclude_object(self, name):
        script = {"script": f"EXCLUDE_OBJECT NAME={name}"}
        self._screen._confirm_send_action(
            None,
            _("Are you sure do you want to exclude the object?") + f"\n\n{name}",
            "printer.gcode.script",
            script
        )

    def draw_graph(self, da, ctx):
        right = da.get_allocated_width() - self.margin_right
        bottom = da.get_allocated_height() - self.margin_bottom
        self.objects = self.printer.get_stat("exclude_object", "objects")

        for obj in self.objects:
            for point in obj["polygon"]:
                # Find min coords
                self.min_x = min(self.min_x, point[0])
                self.min_y = min(self.min_y, point[1])
                # Find max coords
                self.max_x = max(self.max_x, point[0])
                self.max_y = max(self.max_y, point[1])

        # Styling
        ctx.set_source_rgb(.5, .5, .5)  # Grey
        ctx.set_line_width(1)
        ctx.set_font_size(self.font_size)

        # Borders
        ctx.move_to(self.margin_left, self.margin_top)
        # logging.info(f"l:{self.margin_left:.0f} t:{self.margin_top:.0f} r:{right:.0f} b:{bottom:.0f}")
        ctx.line_to(right, self.margin_top)
        ctx.line_to(right, bottom)
        ctx.line_to(self.margin_left, bottom)
        ctx.line_to(self.margin_left, self.margin_top)
        ctx.stroke()

        # Axis labels
        ctx.move_to(0, bottom + self.font_spacing)
        ctx.show_text(f"{self.min_x:.0f},{self.min_y:.0f}")
        ctx.stroke()
        ctx.move_to(right - self.font_spacing * 2, bottom + self.font_spacing)
        ctx.show_text(f"{self.max_x:.0f},{self.min_y:.0f}")
        ctx.stroke()
        ctx.move_to(0, self.font_spacing / 2)
        ctx.show_text(f"{self.min_x:.0f},{self.max_y:.0f}")
        ctx.stroke()

        # middle markers
        midx = (right - self.margin_left) / 2 + self.margin_left
        ctx.set_dash([1, 1])
        ctx.move_to(midx, self.margin_top)
        ctx.line_to(midx, bottom)
        ctx.stroke()
        midy = (self.margin_top - bottom) / 2 + bottom
        ctx.move_to(self.margin_left, midy)
        ctx.line_to(right, midy)
        ctx.stroke()
        ctx.set_dash([1, 0])

        # Draw objects
        for obj in self.objects:
            # change the color depending on the status
            if obj['name'] == self.printer.get_stat("exclude_object", "current_object"):
                ctx.set_source_rgb(1, 0, 0)  # Red
            elif obj['name'] in self.printer.get_stat("exclude_object", "excluded_objects"):
                ctx.set_source_rgb(0, 0, 0)  # Black
            else:
                ctx.set_source_rgb(.5, .5, .5)  # Grey
            for i, point in enumerate(obj["polygon"]):
                # Convert coordinates from bed to screen-graph
                x = self.x_bed_to_graph(da.get_allocated_width(), point[0])
                y = self.y_bed_to_graph(da.get_allocated_height(), point[1])
                if i == 0:
                    ctx.move_to(x, y)
                    continue
                ctx.line_to(x, y)
                # logging.info(f"obj graph: {x=:.0f},{y=:.0f} bed: {point[0]:.0f},{point[1]:.0f}")
            ctx.close_path()
            ctx.fill()
            ctx.stroke()

    def x_bed_to_graph(self, width, bx):
        return (((bx - self.min_x) * (width - self.margin_left - self.margin_right))
                / (self.max_x - self.min_x)) + self.margin_left

    def y_bed_to_graph(self, height, by):
        return ((1 - ((by - self.min_y) / (self.max_y - self.min_y)))
                * (height - self.margin_top - self.margin_bottom)) + self.margin_top



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("Example")
        super().__init__(screen, title)

        # Create gtk items here

        self.content.add(Gtk.Box())



import datetime
import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango
from ks_includes.KlippyGtk import find_widget


class ScreenPanel:
    _screen = None
    _config = None
    _files = None
    _printer = None
    _gtk = None
    ks_printer_cfg = None

    def __init__(self, screen, title, **kwargs):
        self.menu = []
        ScreenPanel._screen = screen
        ScreenPanel._config = screen._config
        ScreenPanel._files = screen.files
        ScreenPanel._printer = screen.printer
        ScreenPanel._gtk = screen.gtk
        self.labels = {}
        self.control = {}
        self.title = title
        self.devices = {}
        self.active_heaters = []
        self.content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, hexpand=True, vexpand=True)
        self.content.get_style_context().add_class("content")
        self._show_heater_power = self._config.get_main_config().getboolean('show_heater_power', False)
        self.bts = self._gtk.bsidescale

        self.update_dialog = None

    def _autoscroll(self, scroll, *args):
        adj = scroll.get_vadjustment()
        adj.set_value(adj.get_upper() - adj.get_page_size())

    def emergency_stop(self, widget):
        if self._config.get_main_config().getboolean('confirm_estop', False):
            self._screen._confirm_send_action(widget, _("Are you sure you want to run Emergency Stop?"),
                                              "printer.emergency_stop")
        else:
            self._screen._ws.klippy.emergency_stop()

    def get_file_image(self, filename, width=None, height=None, small=False):
        if not self._files.has_thumbnail(filename):
            return None
        loc = self._files.get_thumbnail_location(filename, small)
        if loc is None:
            return None
        width = width if width is not None else self._gtk.img_width
        height = height if height is not None else self._gtk.img_height
        if loc[0] == "file":
            return self._gtk.PixbufFromFile(loc[1], width, height)
        if loc[0] == "http":
            return self._gtk.PixbufFromHttp(loc[1], width, height)
        return None

    def menu_item_clicked(self, widget, item):
        panel_args = {}
        if 'name' in item:
            panel_args['title'] = item['name']
        if 'extra' in item:
            panel_args['extra'] = item['extra']
        self._screen.show_panel(item['panel'], **panel_args)

    def load_menu(self, widget, name, title=None):
        logging.info(f"loading menu {name}")
        if f"{name}_menu" not in self.labels:
            logging.error(f"{name} not in labels")
            return

        for child in self.content.get_children():
            self.content.remove(child)

        self.menu.append(f'{name}_menu')
        logging.debug(f"self.menu: {self.menu}")
        self.content.add(self.labels[self.menu[-1]])
        self.content.show_all()
        if title:
            self._screen.base_panel.set_title(f"{self.title} | {title}")

    def unload_menu(self, widget=None):
        if len(self.menu) <= 1 or self.menu[-2] not in self.labels:
            return
        self._screen.base_panel.set_title(self._screen.panels[self._screen._cur_panels[-1]].title)
        self.menu.pop()
        logging.debug(f"self.menu: {self.menu}")
        for child in self.content.get_children():
            self.content.remove(child)
        self.content.add(self.labels[self.menu[-1]])
        self.content.show_all()

    def back(self):
        if len(self.menu) > 1:
            self.unload_menu()
            return True
        return False

    def on_dropdown_change(self, combo, section, option, callback=None):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            model = combo.get_model()
            value = model[tree_iter][1]
            logging.debug(f"[{section}] {option} changed to {value}")
            self._config.set(section, option, value)
            self._config.save_user_config_options()
            if callback is not None:
                callback(value)

    def scale_moved(self, widget, event, section, option):
        logging.debug(f"[{section}] {option} changed to {widget.get_value()}")
        if section not in self._config.get_config().sections():
            self._config.get_config().add_section(section)
        self._config.set(section, option, str(int(widget.get_value())))
        self._config.save_user_config_options()

    def switch_config_option(self, switch, gparam, section, option, callback=None):
        logging.debug(f"[{section}] {option} toggled {switch.get_active()}")
        if section not in self._config.get_config().sections():
            self._config.get_config().add_section(section)
        self._config.set(section, option, "True" if switch.get_active() else "False")
        self._config.save_user_config_options()
        if callback is not None:
            callback(switch.get_active())

    @staticmethod
    def format_time(seconds):
        spc = "\u00A0"  # Non breakable space
        if seconds is None or seconds < 1:
            return "-"
        days = seconds // 86400
        day_units = ngettext("day", "days", days)
        seconds %= 86400
        hours = seconds // 3600
        hour_units = ngettext("hour", "hours", hours)
        seconds %= 3600
        minutes = seconds // 60
        min_units = ngettext("minute", "minutes", minutes)
        seconds %= 60
        sec_units = ngettext("second", "seconds", seconds)
        return f"{f'{days:2.0f}{spc}{day_units}{spc}' if days > 0 else ''}" \
               f"{f'{hours:2.0f}{spc}{hour_units}{spc}' if hours > 0 else ''}" \
               f"{f'{minutes:2.0f}{spc}{min_units}{spc}' if minutes > 0 and days == 0 else ''}" \
               f"{f'{seconds:2.0f}{spc}{sec_units}' if days == 0 and hours == 0 and minutes == 0 else ''}"

    def format_eta(self, total, elapsed):
        if total is None:
            return "-"
        seconds = total - elapsed
        if seconds <= 0:
            return "-"
        days = seconds // 86400
        seconds %= 86400
        hours = seconds // 3600
        seconds %= 3600
        minutes = seconds // 60
        eta = datetime.datetime.now() + datetime.timedelta(days=days, hours=hours, minutes=minutes)
        if self._config.get_main_config().getboolean("24htime", True):
            return f"{self.format_time(total - elapsed)} | {eta:%H:%M} {f' +{days:2.0f}d' if days > 0 else ''}"
        return f"{self.format_time(total - elapsed)} | {eta:%I:%M %p} {f' +{days:2.0f}d' if days > 0 else ''}"

    @staticmethod
    def format_size(size):
        size = float(size)
        suffixes = ["kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
        for i, suffix in enumerate(suffixes, start=2):
            unit = 1024 ** i
            if size < unit:
                return f"{(1024 * size / unit):.1f} {suffix}"

    @staticmethod
    def format_speed(bitrate):
        bitrate = float(bitrate)
        suffixes = ["Kbits/s", "Mbits/s", "Gbits/s", "Tbits/s", "Pbits/s", "Ebits/s", "Zbits/s", "Ybits/s"]
        for i, suffix in enumerate(suffixes, start=1):
            unit = 1000 ** i
            if bitrate < unit:
                return f"{(1000 * bitrate / unit):.0f} {suffix}"

    @staticmethod
    def prettify(name: str):
        name = name.replace("_", " ")
        if name.islower():
            name = name.title()
        return name

    def update_temp(self, dev, temp, target, power, lines=1, digits=1):
        new_label_text = f"{temp or 0:.{digits}f}"
        if self._printer.device_has_target(dev) and target:
            new_label_text += f"/{target:.0f}"
        if dev not in self.devices:
            new_label_text += "°"

        show_power = self._show_heater_power and power
        if show_power:
            new_label_text += f" {power * 100:3.0f}%"

        if dev in self.labels:
            # Job_Status
            find_widget(self.labels[dev], Gtk.Label).set_text(new_label_text)
        elif dev in self.devices:
            # Temperature and Main_Menu
            find_widget(self.devices[dev]["temp"], Gtk.Label).set_text(new_label_text)

    def add_option(self, boxname, opt_array, opt_name, option):
        if option['type'] is None:
            return
        name = Gtk.Label(
            hexpand=True, vexpand=True, halign=Gtk.Align.START, valign=Gtk.Align.CENTER,
            wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR, xalign=0)
        name.set_markup(f"<big><b>{option['name']}</b></big>")

        labels = Gtk.Box(spacing=0, orientation=Gtk.Orientation.VERTICAL, valign=Gtk.Align.CENTER)
        labels.add(name)
        if 'tooltip' in option:
            tooltip = Gtk.Label(
                label=option['tooltip'],
                hexpand=True, vexpand=True, halign=Gtk.Align.START, valign=Gtk.Align.CENTER,
                wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR, xalign=0)
            labels.add(tooltip)

        row_box = Gtk.Box(spacing=5, valign=Gtk.Align.CENTER, hexpand=True, vexpand=False)
        row_box.get_style_context().add_class("frame-item")
        row_box.add(labels)

        setting = {}
        if option['type'] == "binary":
            switch = Gtk.Switch(active=self._config.get_config().getboolean(option['section'], opt_name, fallback=True))
            switch.set_vexpand(False)
            switch.set_valign(Gtk.Align.CENTER)
            switch.connect("notify::active", self.switch_config_option, option['section'], opt_name,
                           option['callback'] if "callback" in option else None)
            row_box.add(switch)
            setting = {opt_name: switch}
        elif option['type'] == "dropdown":
            dropdown = Gtk.ComboBoxText()
            for i, opt in enumerate(option['options']):
                dropdown.append(opt['value'], opt['name'])
                if opt['value'] == self._config.get_config()[option['section']].get(opt_name, option['value']):
                    dropdown.set_active(i)
            dropdown.connect("changed", self.on_dropdown_change, option['section'], opt_name,
                             option['callback'] if "callback" in option else None)
            dropdown.set_entry_text_column(0)
            row_box.add(dropdown)
            setting = {opt_name: dropdown}
        elif option['type'] == "scale":
            row_box.set_orientation(Gtk.Orientation.VERTICAL)
            scale = Gtk.Scale.new_with_range(Gtk.Orientation.HORIZONTAL,
                                             min=option['range'][0], max=option['range'][1], step=option['step'])
            scale.set_hexpand(True)
            scale.set_value(int(self._config.get_config().get(option['section'], opt_name, fallback=option['value'])))
            scale.set_digits(0)
            scale.connect("button-release-event", self.scale_moved, option['section'], opt_name)
            row_box.add(scale)
            setting = {opt_name: scale}
        elif option['type'] == "printer":
            box = Gtk.Box(vexpand=False)
            label = Gtk.Label(f"{option['moonraker_host']}:{option['moonraker_port']}")
            box.add(label)
            row_box.add(box)
        elif option['type'] == "menu":
            open_menu = self._gtk.Button("settings", style="color3")
            open_menu.connect("clicked", self.load_menu, option['menu'], option['name'])
            open_menu.set_hexpand(False)
            open_menu.set_halign(Gtk.Align.END)
            row_box.add(open_menu)
        elif option['type'] == "button":
            select = self._gtk.Button("load", style="color3")
            if "callback" in option:
                select.connect("clicked", option['callback'], option['name'])
            select.set_hexpand(False)
            select.set_halign(Gtk.Align.END)
            row_box.add(select)

        opt_array[opt_name] = {
            "name": option['name'],
            "row": row_box
        }

        opts = sorted(list(opt_array), key=lambda x: opt_array[x]['name'].casefold())
        pos = opts.index(opt_name)

        self.labels[boxname].insert_row(pos)
        self.labels[boxname].attach(opt_array[opt_name]['row'], 0, pos, 1, 1)
        self.labels[boxname].show_all()
        return setting



import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class Keypad(Gtk.Box):
    def __init__(self, screen, ok_cb, cancel_cb, entry_max=5, error_msg=_("Invalid Number")):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self.screen = screen
        self._gtk = screen.gtk
        self.ok_cb = ok_cb
        self.error_msg = error_msg
        self.has_extra_button = False

        # Header: Entry + Close
        self.entry = Gtk.Entry(hexpand=True, xalign=0.5, max_length=entry_max)
        self.entry.connect("activate", self._on_keypad_clicked, "E")

        close_btn = self._gtk.Button('cancel', scale=.66)
        close_btn.set_hexpand(False)
        close_btn.set_vexpand(False)
        close_btn.connect("clicked", cancel_cb)

        self.top_box = Gtk.Box(spacing=5)
        self.top_box.add(self.entry)
        self.top_box.add(close_btn)

        # Grid construction
        numpad = self._build_numpad()

        # Bottom Actions
        self.action_box = Gtk.Box(spacing=5)

        self.ok_btn = self._gtk.Button("complete", style="color1", scale=1.1)
        self.ok_btn.set_vexpand(False)
        self.ok_btn.set_size_request(-1, self._gtk.font_size * (4 if self.screen.vertical_mode else 5))
        self.ok_btn.connect('clicked', self._on_keypad_clicked, "E")

        self.action_box.pack_end(self.ok_btn, True, True, 0)

        # Main Layout
        self.add(self.top_box)
        self.add(numpad)
        self.add(self.action_box)

    def _build_numpad(self):
        grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        grid.get_style_context().add_class('numpad')

        keys = [
            ('1', 'numpad_tleft'), ('2', 'numpad_top'), ('3', 'numpad_tright'),
            ('4', 'numpad_left'), ('5', 'numpad_button'), ('6', 'numpad_right'),
            ('7', 'numpad_left'), ('8', 'numpad_button'), ('9', 'numpad_right'),
            ('B', 'numpad_bleft'), ('0', 'numpad_bottom'), ('.', 'numpad_bright')
        ]

        for i, (label, css_class) in enumerate(keys):
            if label == "B":
                btn = self._gtk.Button("backspace", scale=.66, style="numpad_key")
            else:
                btn = self._gtk.Button(label=label, style="numpad_key")

            btn.connect('clicked', self._on_keypad_clicked, label)
            btn.get_style_context().add_class(css_class)
            grid.attach(btn, i % 3, i // 3, 1, 1)
        return grid

    def add_extra_button(self, callback, icon=None, label=""):
        if self.has_extra_button:
            return
        self.has_extra_button = True
        btn = self._gtk.Button(icon, label, "color2", self._gtk.bsidescale, Gtk.PositionType.LEFT, 1)
        btn.connect("clicked", lambda w: self._handle_extra_action(callback))
        self.action_box.pack_start(btn, True, True, 0)
        btn.set_vexpand(False)
        self.action_box.show_all()
        return btn

    def _handle_extra_action(self, callback):
        value = self.get_value() or 0
        if isinstance(value, (int, float)):
            callback(value)

    def _on_keypad_clicked(self, widget, digit):
        if digit == 'B':
            self.entry.do_backspace(self.entry)
        elif digit == 'E':
            val = self.get_value()
            if val is not None:
                self.ok_cb(val)
                self.clear()
            else:
                self.screen.show_popup_message(self.error_msg)
        else:
            self.entry.do_insert_at_cursor(self.entry, digit)

    def get_value(self):
        try:
            return float(self.entry.get_text())
        except ValueError:
            return None

    def clear(self):
        self.entry.set_text("")



import logging
import mpv

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from contextlib import suppress
from ks_includes.screen_panel import ScreenPanel


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("Camera")
        super().__init__(screen, title)
        self.mpv = None
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        for i, cam in enumerate(self._printer.cameras):
            if not cam["enabled"]:
                continue
            logging.info(cam)
            cam[cam["name"]] = self._gtk.Button(
                image_name="camera", label=cam["name"], style=f"color{i % 4 + 1}",
                scale=self.bts, position=Gtk.PositionType.LEFT, lines=1
            )
            cam[cam["name"]].set_hexpand(True)
            cam[cam["name"]].set_vexpand(True)
            cam[cam["name"]].connect("clicked", self.play, cam)
            box.add(cam[cam["name"]])

        self.scroll = self._gtk.ScrolledWindow()
        self.scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.scroll.add(box)
        self.content.add(self.scroll)
        self.content.show_all()

    def activate(self):
        # if only 1 cam start playing fullscreen
        if len(self._printer.cameras) == 1:
            cam = next(iter(self._printer.cameras))
            if cam['enabled']:
                self.play(None, cam)

    def deactivate(self):
        if self.mpv:
            self.mpv.terminate()
            self.mpv = None

    def play(self, widget, cam):
        url = cam['stream_url']
        if url.startswith('/'):
            logging.info("camera URL is relative")
            endpoint = self._screen.apiclient.endpoint.split(':')
            url = f"{endpoint[0]}:{endpoint[1]}{url}"
        if '/webrtc' in url:
            self._screen.show_popup_message(_('WebRTC is not supported by the backend trying Stream'))
            url = url.replace('/webrtc', '/stream')
        vf_list = []
        if cam["flip_horizontal"]:
            vf_list.append("hflip")
        if cam["flip_vertical"]:
            vf_list.append("vflip")
        if cam["rotation"] != 0:
            vf_list.append(f"rotate:{cam['rotation'] * 3.14159 / 180}")
        logging.info(f"video filters: {vf_list}")

        if self.mpv:
            self.mpv.terminate()
        self.mpv = mpv.MPV(fullscreen=True, log_handler=self.log, vo='gpu,wlshm,xv,x11')

        self.mpv.vf = ','.join(vf_list)

        with suppress(Exception):
            self.mpv.profile = 'sw-fast'

        # LOW LATENCY PLAYBACK
        with suppress(Exception):
            self.mpv.profile = 'low-latency'
        self.mpv.untimed = True
        self.mpv.audio = 'no'

        @self.mpv.on_key_press('MBTN_LEFT' or 'MBTN_LEFT_DBL')
        def clicked():
            self.mpv.quit(0)

        logging.debug(f"Camera URL: {url}")
        self.mpv.play(url)

        try:
            self.mpv.wait_for_playback()
        except mpv.ShutdownError:
            logging.info('Exiting Fullscreen')
        except Exception as e:
            logging.exception(e)
        self.mpv.terminate()
        self.mpv = None
        if len(self._printer.cameras) == 1:
            self._screen._menu_go_back()

    def log(self, loglevel, component, message):
        if (
            'unable to decode' in message  # skip proprietary app fields errors
            or 'No Xvideo support found' in message  # will fall back to other vo automatically
            or 'GBM' in message  # will fall back to other vo automatically
            or 'open TTY for VT control' in message  # not important to notify in the UI
            or 'youtube-dl' in message  # needed for some streams, not relevant for our case
        ):
            return
        if loglevel == 'error':
            self._screen.show_popup_message(f'{message}')
        logging.debug(f'[{loglevel}] {component}: {message}')



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib


class ComboBoxPlus(Gtk.ComboBox):
    """
    A spcial ComboBox that checks if the dropdown was closed too fast and re-opens it
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_open_time = 0
        self.set_focus_on_click(False)
        self.connect("notify::popup-shown", self._on_popup_shown)

    def _on_popup_shown(self, combo, pspec):
        if self.get_property("popup-shown"):
            self._last_open_time = GLib.get_monotonic_time()
            return

        # Microseconds to seconds
        duration = (GLib.get_monotonic_time() - self._last_open_time) / 1000000.0

        if duration < 0.2:
            GLib.idle_add(self._reopen_now, priority=GLib.PRIORITY_HIGH_IDLE)

    def _reopen_now(self):
        self.popup()
        return False



#!/usr/bin/python

import json
import logging
import threading

import gi
import websocket

gi.require_version("Gtk", "3.0")
from gi.repository import GLib
from ks_includes.KlippyGcodes import KlippyGcodes


class KlippyWebsocket(threading.Thread):
    _req_id = 0
    connected = False
    connecting = True
    callback_table = {}
    reconnect_count = 0
    max_retries = 4

    def __init__(self, callback, host, port, api_key, path='', ssl=None):
        threading.Thread.__init__(self)
        self._wst = None
        self.ws_url = None
        self._callback = callback
        self.klippy = MoonrakerApi(self)
        self.ws = None
        self.closing = False
        self.host = host
        self.port = port
        self.path = f"/{path}" if path else ''
        self.ssl = int(self.port) in {443, 7130} if ssl is None else bool(ssl)
        self.header = {"x-api-key": api_key} if api_key else {}
        self.api_key = api_key

    @property
    def _url(self):
        return f"{self.host}:{self.port}{self.path}"

    @property
    def ws_proto(self):
        return "wss" if self.ssl else "ws"

    def initial_connect(self):
        if self.connect() is not False:
            GLib.timeout_add_seconds(10, self.reconnect)

    def reconnect(self):
        if self.reconnect_count > self.max_retries:
            logging.debug("Stopping reconnections")
            self.connecting = False
            GLib.idle_add(self._callback['on_cancel'])
            return False
        return self.connect()

    def connect(self):
        if self.connected:
            logging.debug("Already connected")
            return False
        logging.debug("Attempting to connect")
        self.reconnect_count += 1

        self.ws_url = f"{self.ws_proto}://{self._url}/websocket?token={self.api_key}"
        self.ws = websocket.WebSocketApp(
            self.ws_url,
            on_close=self.on_close,
            on_error=self.on_error,
            on_message=self.on_message,
            on_open=self.on_open,
            header=self.header
        )
        self._wst = threading.Thread(target=self.ws.run_forever, daemon=True)
        try:
            logging.debug("Starting websocket thread")
            self._wst.start()
        except Exception as e:
            logging.debug(f"Error starting web socket {e}")
            return True
        return False

    def close(self):
        logging.debug("Closing websocket")
        self.closing = True
        self.connecting = False
        if self.ws is not None:
            self.ws.keep_running = False
            self.ws.close()

    def on_message(self, *args):
        message = args[1] if len(args) == 2 else args[0]
        response = json.loads(message)
        if "id" in response and response['id'] in self.callback_table:
            args = (response,
                    self.callback_table[response['id']][1],
                    self.callback_table[response['id']][2],
                    *self.callback_table[response['id']][3])
            GLib.idle_add(self.callback_table[response['id']][0], *args, priority=GLib.PRIORITY_HIGH_IDLE)
            self.callback_table.pop(response['id'])
            return

        if "method" in response and "on_message" in self._callback:
            args = (response['method'], response['params'][0] if "params" in response else {})
            GLib.idle_add(self._callback['on_message'], *args, priority=GLib.PRIORITY_HIGH_IDLE)
        if self.closing:
            timer = threading.Timer(2, self.ws.close)
            timer.start()
        return

    def send_method(self, method, params=None, callback=None, *args):
        if not self.connected or self.closing:
            return False
        if params is None:
            params = {}

        self._req_id += 1
        if callback is not None:
            self.callback_table[self._req_id] = [callback, method, params, [*args]]

        data = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
            "id": self._req_id
        }
        self.ws.send(json.dumps(data))
        return True

    def on_open(self, *args):
        logging.info("Moonraker Websocket Open")
        self.connected = True
        self.connecting = False
        self.reconnect_count = 0
        if "on_connect" in self._callback:
            GLib.idle_add(self._callback['on_connect'], priority=GLib.PRIORITY_HIGH_IDLE)

    def on_close(self, *args):
        # args: ws, status, message
        # sometimes ws is not passed due to bugs
        if len(args) == 3:
            status = args[1]
            message = args[2]
        else:
            status = args[0]
            message = args[1]
        if message is not None:
            logging.info(f"{status} {message}")
        if not self.connected:
            logging.debug("Connection already closed")
            return
        if "on_close" in self._callback:
            GLib.idle_add(self._callback['on_close'], priority=GLib.PRIORITY_HIGH_IDLE)
        logging.info("Moonraker Websocket Closed")
        self.connected = False

    @staticmethod
    def on_error(*args):
        error = args[1] if len(args) == 2 else args[0]
        logging.debug(f"Websocket error: {error}")


class MoonrakerApi:
    def __init__(self, ws):
        self._ws = ws

    def emergency_stop(self):
        logging.info("Sending printer.emergency_stop")
        return self._ws.send_method(
            "printer.emergency_stop"
        )

    def gcode_script(self, script, callback=None, *args):
        logging.debug(f"Sending printer.gcode.script: {script}")
        return self._ws.send_method(
            "printer.gcode.script",
            {"script": script},
            callback,
            *args
        )

    def get_file_dir(self, path='gcodes', callback=None, *args):
        logging.debug(f"Sending server.files.directory {path}")
        return self._ws.send_method(
            "server.files.list",
            {"path": path},
            callback,
            *args
        )

    def get_file_list(self, callback=None, *args):
        logging.debug("Sending server.files.list")
        return self._ws.send_method(
            "server.files.list",
            {},
            callback,
            *args
        )

    def get_dir_info(self, callback=None, directory='gcodes', *args):
        logging.debug(f"Sending server.files.get_directory  {directory}")
        return self._ws.send_method(
            "server.files.get_directory",
            {"path": directory},
            callback,
            *args
        )

    def get_file_metadata(self, filename, callback=None, *args):
        return self._ws.send_method(
            "server.files.metadata",
            {"filename": filename},
            callback,
            *args
        )

    def object_subscription(self, updates):
        logging.debug("Sending printer.objects.subscribe")
        return self._ws.send_method(
            "printer.objects.subscribe",
            updates
        )

    def power_device_off(self, device, callback=None, *args):
        logging.debug(f"Sending machine.device_power.off: {device}")
        return self._ws.send_method(
            "machine.device_power.off",
            {device: False},
            callback,
            *args
        )

    def power_device_on(self, device, callback=None, *args):
        logging.debug("Sending machine.device_power.on {device}")
        return self._ws.send_method(
            "machine.device_power.on",
            {device: False},
            callback,
            *args
        )

    def print_cancel(self, callback=None, *args):
        logging.debug("Sending printer.print.cancel")
        return self._ws.send_method(
            "printer.print.cancel",
            {},
            callback,
            *args
        )

    def print_pause(self, callback=None, *args):
        logging.debug("Sending printer.print.pause")
        return self._ws.send_method(
            "printer.print.pause",
            {},
            callback,
            *args
        )

    def print_resume(self, callback=None, *args):
        logging.debug("Sending printer.print.resume")
        return self._ws.send_method(
            "printer.print.resume",
            {},
            callback,
            *args
        )

    def print_start(self, filename, callback=None, *args):
        logging.debug("Sending printer.print.start")
        return self._ws.send_method(
            "printer.print.start",
            {
                "filename": filename
            },
            callback,
            *args
        )

    def set_bed_temp(self, target, callback=None, *args):
        logging.debug(f"Sending set_bed_temp: {KlippyGcodes.set_bed_temp(target)}")
        return self._ws.send_method(
            "printer.gcode.script",
            {
                "script": KlippyGcodes.set_bed_temp(target)
            },
            callback,
            *args
        )

    def set_heater_temp(self, heater, target, callback=None, *args):
        logging.debug(f"Sending heater {heater} to temp: {target}")
        return self._ws.send_method(
            "printer.gcode.script",
            {
                "script": KlippyGcodes.set_heater_temp(heater, target)
            },
            callback,
            *args
        )

    def set_temp_fan_temp(self, temp_fan, target, callback=None, *args):
        logging.debug(f"Sending temperature fan {temp_fan} to temp: {target}")
        return self._ws.send_method(
            "printer.gcode.script",
            {
                "script": KlippyGcodes.set_temp_fan_temp(temp_fan, target)
            },
            callback,
            *args
        )

    def set_tool_temp(self, tool, target, callback=None, *args):
        logging.debug(f"Sending set_tool_temp: {KlippyGcodes.set_ext_temp(target, tool)}")
        return self._ws.send_method(
            "printer.gcode.script",
            {
                "script": KlippyGcodes.set_ext_temp(target, tool)
            },
            callback,
            *args
        )

    def restart(self):
        logging.debug("Sending printer.restart")
        return self._ws.send_method(
            "printer.restart"
        )

    def restart_firmware(self):
        logging.debug("Sending printer.firmware_restart")
        return self._ws.send_method(
            "printer.firmware_restart"
        )

    def identify_client(self, version, api_key):
        logging.debug("Sending server.connection.identify")
        return self._ws.send_method(
            "server.connection.identify",
            {
                "client_name": "KlipperScreen",
                "version": f"{version}",
                "type": "display",
                "url": "https://github.com/KlipperScreen/KlipperScreen",
                "api_key": f"{api_key}"
            },
        )



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from math import pi
from ks_includes.KlippyGcodes import KlippyGcodes
from ks_includes.screen_panel import ScreenPanel


def rgb_to_hex(color):
    hex_color = '#'
    for value in color:
        int_value = round(value * 255)
        hex_color += hex(int_value)[2:].zfill(2)
    return hex_color.upper()


def rgbw_to_rgb(color):
    # The idea here is to use the white channel as a saturation control
    # The white channel 'washes' the color
    return (
        [color[3] for i in range(3)]  # Special case of only white channel
        if color[0] == 0 and color[1] == 0 and color[2] == 0
        else [color[i] + (1 - color[i]) * color[3] / 3 for i in range(3)]
    )


class Panel(ScreenPanel):

    def __init__(self, screen, title):
        title = title or _("Leds")
        super().__init__(screen, title)
        self.da_size = self._gtk.img_scale * 2
        self.preview = ColorPreviewArea(size=self.da_size)
        self.preview.set_size_request(-1, self.da_size * 2)
        self.preview_label = Gtk.Label()
        self.preset_list = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        self.color_data = [0, 0, 0, 0]
        self.color_order = 'RGBW'
        self.presets = {"off": [0.0, 0.0, 0.0, 0.0]}
        self.scales = {}
        self.buttons = []
        self.leds = self._printer.get_leds()
        self.current_led = self.leds[0] if len(self.leds) == 1 else None
        self.open_selector(None, self.current_led)

    def color_available(self, idx):
        return (
            (idx == 0 and 'R' in self.color_order)
            or (idx == 1 and 'G' in self.color_order)
            or (idx == 2 and 'B' in self.color_order)
            or (idx == 3 and 'W' in self.color_order)
        )

    def activate(self):
        if self.current_led is not None:
            self.set_title(f"{self.current_led}")

    def set_title(self, title):
        self._screen.base_panel.set_title(self.prettify(title))

    def back(self):
        if len(self.leds) > 1 and self.current_led:
            self.set_title(self._screen.panels[self._screen._cur_panels[-1]].title)
            self.open_selector(led=None)
            return True
        return False

    def open_selector(self, widget=None, led=None):
        for child in self.content.get_children():
            self.content.remove(child)
        if led is None:
            self.content.add(self.led_selector())
        else:
            self.content.add(self.color_selector(led))
        self.content.show_all()

    def led_selector(self):
        self.current_led = None
        columns = 3 if self._screen.vertical_mode else 4
        grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        for i, led in enumerate(self.leds):
            name = led.split()[1] if len(led.split()) > 1 else led
            button = self._gtk.Button(None, name.upper(), style=f"color{(i % 4) + 1}")
            button.connect("clicked", self.open_selector, led)
            grid.attach(button, (i % columns), int(i / columns), 1, 1)
        scroll = self._gtk.ScrolledWindow()
        scroll.add(grid)
        return scroll

    def color_selector(self, led):
        logging.info(led)
        self.current_led = led
        self.set_title(f"{self.current_led}")
        grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        self.color_order = self._printer.get_led_color_order(led)
        if self.color_order is None:
            logging.error("Error: Color order is None")
            self.back()
            return
        on = [1 if self.color_available(i) else 0 for i in range(4)]
        self.presets["on"] = on
        scale_grid = Gtk.Grid(row_homogeneous=True, column_homogeneous=True)
        for idx, col_value in enumerate(self.color_data):
            if not self.color_available(idx):
                continue
            color = [0, 0, 0, 0]
            color[idx] = 1
            button = self._gtk.Button()
            preview = ColorPreviewArea(size=self.da_size)
            preview.set_color(color)
            button.set_image(preview)
            button.connect("clicked", self.apply_preset, color)
            button.set_hexpand(False)
            scale = Gtk.Scale.new_with_range(Gtk.Orientation.HORIZONTAL, min=0, max=255, step=1)
            scale.set_value(round(col_value * 255))
            scale.set_digits(0)
            scale.set_hexpand(True)
            scale.set_has_origin(True)
            scale.get_style_context().add_class("fan_slider")
            scale.connect("button-release-event", self.apply_scales)
            scale.connect("value_changed", self.update_preview_label)
            self.scales[idx] = scale
            scale_grid.attach(button, 0, idx, 1, 1)
            scale_grid.attach(scale, 1, idx, 3, 1)
        grid.attach(scale_grid, 0, 0, 3, 1)

        columns = 3 if self._screen.vertical_mode else 2
        data_misc = self._screen.apiclient.send_request(
            "server/database/item?namespace=mainsail&key=miscellaneous.entries")
        if data_misc:
            presets_data = data_misc['value'][next(iter(data_misc["value"]))]['presets']
            if presets_data:
                self.presets.update(self.parse_presets(presets_data))
        for i, key in enumerate(self.presets):
            logging.info(f'Adding preset: {key}')
            preview = ColorPreviewArea(size=self.da_size)
            preview.set_color(self.presets[key])
            button = self._gtk.Button()
            button.set_image(preview)
            button.connect("clicked", self.apply_preset, self.presets[key])
            self.preset_list.attach(button, i % columns, int(i / columns) + 1, 1, 1)

        scroll = self._gtk.ScrolledWindow()
        scroll.add(self.preset_list)
        preview_box = Gtk.Box(homogeneous=True)
        preview_box.add(self.preview_label)
        preview_box.add(self.preview)
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.add(preview_box)
        box.add(scroll)
        if self._screen.vertical_mode:
            grid.attach(box, 0, 1, 3, 1)
        else:
            grid.attach(box, 3, 0, 2, 1)
        return grid

    def update_preview_label(self, args):
        self.preview_label.set_label(rgb_to_hex(rgbw_to_rgb(self.color_data)))

    def process_update(self, action, data):
        if action != 'notify_status_update':
            return
        if self.current_led in data and "color_data" in data[self.current_led]:
            self.update_scales(data[self.current_led]["color_data"][0])
        self.preview.set_color(self.color_data)

    def update_scales(self, color_data):
        for idx in self.scales:
            self.scales[idx].set_value(int(color_data[idx] * 255))
        self.color_data = color_data

    def update_color_data(self):
        for idx in self.scales:
            self.color_data[idx] = self.scales[idx].get_value() / 255
        self.preview.set_color(self.color_data)

    def apply_preset(self, widget, color_data):
        self.update_scales(color_data)
        self.apply_scales()

    def apply_scales(self, *args):
        self.update_color_data()
        self.set_led_color(self.color_data)

    def set_led_color(self, color_data):
        name = self.current_led.split()[1] if len(self.current_led.split()) > 1 else self.current_led
        self._screen._send_action(None, "printer.gcode.script",
                                  {"script": KlippyGcodes.set_led_color(name, color_data)})

    @staticmethod
    def parse_presets(presets_data) -> {}:
        parsed = {}
        for i, preset in enumerate(presets_data.values()):
            name = i if preset["name"] == '' else preset["name"].lower()
            parsed[name] = []
            for color in ["red", "green", "blue", "white"]:
                if color not in preset or preset[color] is None:
                    parsed[name].append(0)
                    continue
                parsed[name].append(preset[color] / 255)
        return parsed


class ColorPreviewArea(Gtk.DrawingArea):
    color = [0, 0, 0]

    def __init__(self, size=-1):
        super().__init__(width_request=size, height_request=size)
        self.connect("draw", self.on_draw)

    def set_color(self, value):
        logging.debug(f"color: {value}")
        self.color = rgbw_to_rgb(value)
        self.queue_draw()

    def on_draw(self, da, ctx):
        ctx.set_source_rgb(*self.color)
        # Set the size of the rectangle
        width = da.get_allocated_width() * .9
        height = da.get_allocated_height() * .9
        x = da.get_allocated_width() * .05
        # Set the radius of the corners
        radius = width / 2 * 0.2
        ctx.arc(x + radius, radius, radius, pi, 3 * pi / 2)
        ctx.arc(x + width - radius, radius, radius, 3 * pi / 2, 0)
        ctx.arc(x + width - radius, height - radius, radius, 0, pi / 2)
        ctx.arc(x + radius, height - radius, radius, pi / 2, pi)
        ctx.close_path()
        ctx.fill()



import logging

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from ks_includes.screen_panel import ScreenPanel
from ks_includes.widgets.autogrid import AutoGrid
from ks_includes.KlippyGtk import find_widget


class Panel(ScreenPanel):
    def __init__(self, screen, title):
        title = title or _("Printer Select")
        super().__init__(screen, title)
        printers = self._config.get_printers()

        self.printer_buttons = []
        scale = 3
        image_scale = scale * self._gtk.img_scale
        if len(printers) == 2:
            image_scale *= 2
        elif len(printers) == 3:
            image_scale *= 1.6
        elif 4 <= len(printers) <= 6:
            image_scale *= 1.4
        for i, printer in enumerate(printers):
            name = list(printer)[0]
            self.labels[name] = self._gtk.Button("printer", name, f"color{1 + i % 4}", scale=scale)
            pixbuf = self._gtk.PixbufFromIcon(f"../../printers/{name}", image_scale, image_scale)
            if pixbuf is not None:
                image = find_widget(self.labels[name], Gtk.Image)
                image.set_from_pixbuf(pixbuf)
            self.labels[name].connect("clicked", self.connect_printer, name)
            self.printer_buttons.append(self.labels[name])
        grid = AutoGrid(self.printer_buttons, vertical=self._screen.vertical_mode)

        scroll = self._gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scroll.add(grid)
        self.content.add(scroll)

    def connect_printer(self, widget, name):
        self._screen.connect_printer(name)

    def activate(self):
        if self._screen._ws and self._screen._ws.connected:
            self._screen.close_websocket()
            logging.debug("Waiting for disconnect")
            self._screen.base_panel.set_title(_("Please wait"))
            b: Gtk.Button
            for b in self.printer_buttons:
                b.set_sensitive(False)

    def disconnected_callback(self):
        logging.debug("Disconnected. Enabling buttons")
        self._screen.base_panel.set_title(_("Printer Select"))
        for b in self.printer_buttons:
            b.set_sensitive(True)
